#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.common import rpc
from esi_leap.tests import base

import mock

from oslo_config import cfg
from oslo_context import context as ctx
import oslo_messaging as messaging

CONF = cfg.CONF


# TestUtils borrowed from Ironic
class TestUtils(base.TestCase):

    @mock.patch.object(messaging, 'Notifier', autospec=True)
    @mock.patch.object(messaging, 'JsonPayloadSerializer', autospec=True)
    @mock.patch.object(messaging, 'get_notification_transport', autospec=True)
    def test_init_globals_notifications_disabled(self,
                                                 mock_get_notification,
                                                 mock_json_serializer,
                                                 mock_notifier):
        self._test_init_globals(False,
                                mock_get_notification, mock_json_serializer,
                                mock_notifier)

    @mock.patch.object(messaging, 'Notifier', autospec=True)
    @mock.patch.object(messaging, 'JsonPayloadSerializer', autospec=True)
    @mock.patch.object(messaging, 'get_notification_transport', autospec=True)
    def test_init_globals_notifications_enabled(self,
                                                mock_get_notification,
                                                mock_json_serializer,
                                                mock_notifier):
        self.config(notification_level='debug', group='notification')
        self._test_init_globals(True,
                                mock_get_notification, mock_json_serializer,
                                mock_notifier)

    def _test_init_globals(
        self, notifications_enabled,
        mock_get_notification, mock_json_serializer,
        mock_notifier,
        versioned_notifications_topics=(
            ['esi_leap_versioned_notifications'])):

        rpc.NOTIFICATION_TRANSPORT = None
        rpc.VERSIONED_NOTIFIER = None
        mock_request_serializer = mock.Mock()
        mock_request_serializer.return_value = mock.Mock()
        rpc.RequestContextSerializer = mock_request_serializer

        mock_notifier.return_value = mock.Mock()

        rpc.init(CONF)

        self.assertEqual(mock_get_notification.return_value,
                         rpc.NOTIFICATION_TRANSPORT)
        self.assertTrue(mock_json_serializer.called)

        if not notifications_enabled:
            mock_notifier.assert_any_call(
                rpc.NOTIFICATION_TRANSPORT,
                serializer=mock_request_serializer.return_value,
                driver='noop')
        else:
            mock_notifier.assert_any_call(
                rpc.NOTIFICATION_TRANSPORT,
                serializer=mock_request_serializer.return_value,
                topics=versioned_notifications_topics)

        self.assertEqual(mock_notifier.return_value, rpc.VERSIONED_NOTIFIER)

    def test_get_versioned_notifier(self):
        rpc.VERSIONED_NOTIFIER = mock.Mock(autospec=True)
        rpc.get_versioned_notifier(publisher_id='a_great_publisher')
        rpc.VERSIONED_NOTIFIER.prepare.assert_called_once_with(
            publisher_id='a_great_publisher')

    def test_get_versioned_notifier_no_publisher_id(self):
        rpc.VERSIONED_NOTIFIER = mock.Mock()
        self.assertRaises(AssertionError,
                          rpc.get_versioned_notifier, publisher_id=None)

    def test_get_versioned_notifier_no_notifier(self):
        rpc.VERSIONED_NOTIFIER = None
        self.assertRaises(
            AssertionError,
            rpc.get_versioned_notifier, publisher_id='a_great_publisher')


class TestRequestContextSerializer(base.TestCase):

    def setUp(self):
        super(TestRequestContextSerializer, self).setUp()

        self.mock_serializer = mock.MagicMock()
        self.serializer = rpc.RequestContextSerializer(self.mock_serializer)
        self.context = ctx.RequestContext()
        self.entity = {'foo': 'bar'}

    def test_serialize_entity(self):
        self.serializer.serialize_entity(self.context, self.entity)
        self.mock_serializer.serialize_entity.assert_called_with(
            self.context, self.entity)

    def test_serialize_entity_empty_base(self):
        # NOTE(viktors): Return False for check `if self.serializer._base:`
        bool_args = {'__bool__': lambda *args: False,
                     '__nonzero__': lambda *args: False}
        self.mock_serializer.configure_mock(**bool_args)

        entity = self.serializer.serialize_entity(self.context, self.entity)
        self.assertFalse(self.mock_serializer.serialize_entity.called)
        # If self.serializer._base is empty, return entity directly
        self.assertEqual(self.entity, entity)

    def test_deserialize_entity(self):
        self.serializer.deserialize_entity(self.context, self.entity)
        self.mock_serializer.deserialize_entity.assert_called_with(
            self.context, self.entity)

    def test_deserialize_entity_empty_base(self):
        # NOTE(viktors): Return False for check `if self.serializer._base:`
        bool_args = {'__bool__': lambda *args: False,
                     '__nonzero__': lambda *args: False}
        self.mock_serializer.configure_mock(**bool_args)

        entity = self.serializer.deserialize_entity(self.context, self.entity)
        self.assertFalse(self.mock_serializer.serialize_entity.called)
        self.assertEqual(self.entity, entity)

    def test_serialize_context(self):
        serialize_values = self.serializer.serialize_context(self.context)

        self.assertEqual(self.context.to_dict(), serialize_values)

    def test_deserialize_context(self):
        serialize_values = self.context.to_dict()
        new_context = self.serializer.deserialize_context(serialize_values)
        self.assertEqual(serialize_values, new_context.to_dict())
        self.assertIsInstance(new_context, ctx.RequestContext)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import mock

from esi_leap.common import notification_utils as notif_utils
from esi_leap.objects import fields
from esi_leap.objects import lease as lease_obj
from esi_leap.resource_objects.test_node import TestNode
from esi_leap.tests import base as tests_base


class NotifyTestCase(tests_base.TestCase):

    def setUp(self):
        super(NotifyTestCase, self).setUp()
        self.lease_notify_mock = mock.Mock()
        self.lease_notify_mock.__name__ = 'LeaseCRUDNotification'
        self.crud_notify_obj = {
            'lease': (self.lease_notify_mock,
                      lease_obj.LeaseCRUDPayload),
        }
        self.lease = lease_obj.Lease(
            id='12345',
            name='test_lease',
            start_time=datetime.datetime(2016, 7, 16, 19, 20, 30),
            end_time=datetime.datetime(2016, 8, 16, 19, 20, 30),
            fulfill_time=datetime.datetime(2016, 7, 16, 19, 21, 30),
            expire_time=datetime.datetime(2016, 8, 16, 19, 21, 30),
            uuid='13921c8d-ce11-4b6d-99ed-10e19d184e5f',
            resource_type='test_node',
            resource_uuid='111',
            project_id='lesseeid',
            owner_id='ownerid',
            parent_lease_uuid=None,
            offer_uuid=None,
            properties=None,
            status='created',
            purpose=None,
        )
        self.node = TestNode('test-node', '12345')

    def test_emit_notification(self):
        self.config(host='fake-host')
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_notification(self.context,
                                       self.lease,
                                       'fulfill', test_level,
                                       test_status,
                                       self.crud_notify_obj,
                                       node=self.node)
        init_kwargs = self.lease_notify_mock.call_args[1]
        publisher = init_kwargs['publisher']
        event_type = init_kwargs['event_type']
        payload = init_kwargs['payload']
        level = init_kwargs['level']
        self.assertEqual('fake-host', publisher.host)
        self.assertEqual('esi-leap-manager', publisher.service)
        self.assertEqual('fulfill', event_type.action)
        self.assertEqual('lease', event_type.object)
        self.assertEqual(test_status, event_type.status)
        self.assertEqual(test_level, level)
        self.assertEqual(self.lease.name, payload.name)
        self.assertEqual(self.lease.uuid, payload.uuid)
        self.assertEqual(self.node.node_name, payload.node_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from keystoneclient import client as keystone_client
from oslo_utils import uuidutils

from esi_leap.common import exception
import esi_leap.conf


CONF = esi_leap.conf.CONF
_cached_keystone_client = None
_cached_project_list = None


def get_keystone_client():
    global _cached_keystone_client
    if _cached_keystone_client is not None:
        return _cached_keystone_client

    auth_plugin = ks_loading.load_auth_from_conf_options(CONF, 'keystone')
    sess = ks_loading.load_session_from_conf_options(CONF, 'keystone',
                                                     auth=auth_plugin)
    cli = keystone_client.Client(session=sess)
    _cached_keystone_client = cli

    return cli


def get_parent_project_id_tree(project_id):
    ks_client = get_keystone_client()
    project = ks_client.projects.get(project_id)
    project_ids = [project.id]
    while project.parent_id is not None:
        project = ks_client.projects.get(project.parent_id)
        project_ids.append(project.id)
    return project_ids


def get_project_uuid_from_ident(project_ident):
    if uuidutils.is_uuid_like(project_ident):
        return project_ident
    else:
        projects = get_keystone_client().projects.list(name=project_ident)
        if len(projects) > 0:
            # projects have unique names
            return projects[0].id
        raise exception.ProjectNoSuchName(name=project_ident)


def get_project_list():
    return get_keystone_client().projects.list()


def get_project_name(project_id, project_list=None):
    if project_id:
        if project_list is None:
            project = get_keystone_client().projects.get(project_id)
        else:
            project = next((p for p in project_list
                            if getattr(p, 'id') == project_id),
                           None)
        return project.name if project else ''
    else:
        return ''



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import concurrent.futures
import datetime
import http.client as http_client
from oslo_utils import uuidutils
import pecan
from pecan import rest
import wsme
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from esi_leap.api.controllers import base
from esi_leap.api.controllers import types
from esi_leap.api.controllers.v1 import utils
from esi_leap.common import constants
from esi_leap.common import exception
from esi_leap.common import ironic
from esi_leap.common import keystone
from esi_leap.common import statuses
import esi_leap.conf
from esi_leap.objects import lease as lease_obj
from esi_leap.resource_objects import get_resource_object

CONF = esi_leap.conf.CONF


class Lease(base.ESILEAPBase):

    name = wsme.wsattr(wtypes.text)
    uuid = wsme.wsattr(wtypes.text, readonly=True)
    project_id = wsme.wsattr(wtypes.text)
    project = wsme.wsattr(wtypes.text, readonly=True)
    owner_id = wsme.wsattr(wtypes.text, readonly=True)
    owner = wsme.wsattr(wtypes.text, readonly=True)
    resource_type = wsme.wsattr(wtypes.text)
    resource_uuid = wsme.wsattr(wtypes.text)
    resource_class = wsme.wsattr(wtypes.text)
    resource = wsme.wsattr(wtypes.text, readonly=True)
    start_time = wsme.wsattr(datetime.datetime)
    fulfill_time = wsme.wsattr(datetime.datetime, readonly=True)
    expire_time = wsme.wsattr(datetime.datetime, readonly=True)
    end_time = wsme.wsattr(datetime.datetime)
    status = wsme.wsattr(wtypes.text, readonly=True)
    properties = {wtypes.text: types.jsontype}
    purpose = wsme.wsattr(wtypes.text)
    offer_uuid = wsme.wsattr(wtypes.text, readonly=True)
    parent_lease_uuid = wsme.wsattr(wtypes.text, readonly=True)

    def __init__(self, **kwargs):
        self.fields = lease_obj.Lease.fields
        for field in self.fields:
            setattr(self, field, kwargs.get(field, wtypes.Unset))

        for attr in ('project', 'owner', 'resource', 'resource_class'):
            setattr(self, attr, kwargs.get(attr, wtypes.Unset))


class LeaseCollection(types.Collection):
    leases = [Lease]

    def __init__(self, **kwargs):
        self._type = 'leases'


class LeasesController(rest.RestController):

    @wsme_pecan.wsexpose(Lease, wtypes.text)
    def get_one(self, lease_id):
        request = pecan.request.context

        lease = utils.check_lease_policy_and_retrieve(
            request, 'esi_leap:lease:get', lease_id)

        return Lease(**utils.lease_get_dict_with_added_info(lease))

    @wsme_pecan.wsexpose(LeaseCollection, wtypes.text,
                         datetime.datetime, datetime.datetime, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text)
    def get_all(self, project_id=None, start_time=None, end_time=None,
                status=None, offer_uuid=None, view=None, owner_id=None,
                resource_type=None, resource_uuid=None, resource_class=None):
        request = pecan.request.context
        cdict = request.to_policy_values()

        if project_id is not None:
            project_id = keystone.get_project_uuid_from_ident(project_id)

        if owner_id is not None:
            owner_id = keystone.get_project_uuid_from_ident(owner_id)

        if resource_uuid is not None:
            if resource_type is None:
                resource_type = CONF.api.default_resource_type

            resource = get_resource_object(resource_type, resource_uuid)
            resource_uuid = resource.get_uuid()

        filters = LeasesController._lease_get_all_authorize_filters(
            cdict, project_id=project_id, owner_id=owner_id,
            start_time=start_time, end_time=end_time,
            status=status, offer_uuid=offer_uuid, view=view,
            resource_type=resource_type, resource_uuid=resource_uuid)

        lease_collection = LeaseCollection()
        leases = lease_obj.Lease.get_all(filters, request)

        lease_collection.leases = []

        if len(leases) > 0:
            project_list = None
            node_list = None

            with concurrent.futures.ThreadPoolExecutor() as executor:
                f1 = executor.submit(ironic.get_node_list)
                f2 = executor.submit(keystone.get_project_list)
                node_list = f1.result()
                project_list = f2.result()

            leases_with_added_info = [
                Lease(**utils.lease_get_dict_with_added_info(l, project_list,
                                                             node_list))
                for l in leases]
            if resource_class:
                lease_collection.leases = [
                    l for l in leases_with_added_info
                    if l.resource_class == resource_class]
            else:
                lease_collection.leases = leases_with_added_info

        return lease_collection

    @wsme_pecan.wsexpose(Lease, body=Lease, status_code=http_client.CREATED)
    def post(self, new_lease):
        request = pecan.request.context
        cdict = request.to_policy_values()
        utils.policy_authorize('esi_leap:lease:create', cdict, cdict)

        lease_dict = new_lease.to_dict()
        lease_dict['owner_id'] = request.project_id
        lease_dict['uuid'] = uuidutils.generate_uuid()
        if 'resource_type' not in lease_dict:
            lease_dict['resource_type'] = CONF.api.default_resource_type
        resource = get_resource_object(lease_dict['resource_type'],
                                       lease_dict['resource_uuid'])
        lease_dict['resource_uuid'] = resource.get_uuid()

        if 'project_id' in lease_dict:
            lease_dict['project_id'] = keystone.get_project_uuid_from_ident(
                lease_dict['project_id'])

        if 'start_time' not in lease_dict:
            lease_dict['start_time'] = datetime.datetime.now()

        if 'end_time' not in lease_dict:
            lease_dict['end_time'] = lease_dict['start_time'] + \
                datetime.timedelta(days=CONF.api.default_lease_time)
        else:
            utils.check_lease_length(cdict,
                                     lease_dict['start_time'],
                                     lease_dict['end_time'],
                                     CONF.api.max_lease_time)

        try:
            utils.check_resource_admin(cdict, resource, request.project_id)
        except exception.HTTPResourceForbidden:
            parent_lease_uuid = utils.check_resource_lease_admin(
                cdict,
                resource,
                request.project_id,
                lease_dict.get('start_time'),
                lease_dict.get('end_time'))
            if parent_lease_uuid is None:
                raise
            lease_dict['parent_lease_uuid'] = parent_lease_uuid

        lease = lease_obj.Lease(**lease_dict)
        lease.create(request)
        return Lease(**utils.lease_get_dict_with_added_info(lease))

    @wsme_pecan.wsexpose(Lease, wtypes.text)
    def delete(self, lease_id):
        request = pecan.request.context

        lease = utils.check_lease_policy_and_retrieve(
            request, 'esi_leap:lease:get', lease_id,
            statuses.LEASE_CAN_DELETE)

        lease.cancel(request)

    @staticmethod
    def _lease_get_all_authorize_filters(cdict,
                                         start_time=None, end_time=None,
                                         status=None, offer_uuid=None,
                                         project_id=None, view=None,
                                         owner_id=None, resource_type=None,
                                         resource_uuid=None):

        if status is not None:
            status = [status] if status != 'any' else None
        else:
            status = [statuses.CREATED, statuses.ACTIVE, statuses.ERROR,
                      statuses.WAIT_CANCEL, statuses.WAIT_EXPIRE,
                      statuses.WAIT_FULFILL]

        filters = {
            'status': status,
            'offer_uuid': offer_uuid,
            'start_time': start_time,
            'end_time': end_time,
            'resource_type': resource_type,
            'resource_uuid': resource_uuid,
            'time_filter_type': constants.WITHIN_TIME_FILTER,
        }

        if view == 'all':
            utils.policy_authorize('esi_leap:lease:lease_admin', cdict, cdict)
            filters['owner_id'] = owner_id
            filters['project_id'] = project_id
        else:
            utils.policy_authorize('esi_leap:lease:get_all', cdict, cdict)

            if owner_id:
                if cdict['project_id'] != owner_id:
                    utils.policy_authorize('esi_leap:lease:lease_admin', cdict,
                                           cdict)

                filters['owner_id'] = owner_id
                filters['project_id'] = project_id
            else:
                if project_id is None:
                    project_id = cdict['project_id']
                    filters['project_or_owner_id'] = project_id
                else:
                    if project_id != cdict['project_id']:
                        utils.policy_authorize('esi_leap:lease:lease_admin',
                                               cdict, cdict)
                    filters['project_id'] = project_id

        # either both are defined or both are None
        if bool(start_time) != bool(end_time):
            raise exception.InvalidTimeAPICommand(resource='a lease',
                                                  start_time=str(start_time),
                                                  end_time=str(end_time))
        if (start_time or end_time) and (end_time <= start_time):
            raise exception.InvalidTimeAPICommand(resource='a lease',
                                                  start_time=str(start_time),
                                                  end_time=str(end_time))

        # unpack iterator to tuple so we can use 'del'
        for k, v in tuple(filters.items()):
            if v is None:
                del filters[k]

        return filters



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from oslo_policy import policy

import esi_leap.conf

CONF = esi_leap.conf.CONF
_ENFORCER = None

default_policies = [
    policy.RuleDefault('is_admin',
                       'role:admin or role:esi_leap_admin',
                       description='Full read/write API access'),
    policy.RuleDefault('is_owner',
                       'role:owner or role:esi_leap_owner',
                       description='Owner API access'),
    policy.RuleDefault('is_lessee',
                       'role:lessee or role:esi_leap_lessee',
                       description='Lessee API access'),
    policy.RuleDefault('is_offer_owner',
                       'project_id:%(offer.project_id)s',
                       description='Owner of offer'),
    policy.RuleDefault('is_lease_owner',
                       'project_id:%(lease.owner_id)s',
                       description='Owner of lease'),
    policy.RuleDefault('is_lease_lessee',
                       'project_id:%(lease.project_id)s',
                       description='Lessee of lease'),
]

lease_policies = [
    policy.DocumentedRuleDefault(
        'esi_leap:lease:lease_admin',
        'rule:is_admin',
        'Complete permissions over leases',
        [{'path': '/leases', 'method': 'POST'},
         {'path': '/leases', 'method': 'GET'},
         {'path': '/leases/{lease_ident}', 'method': 'GET'},
         {'path': '/leases/{lease_ident}', 'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:lease:create',
        'rule:is_admin or rule:is_owner',
        'Create lease',
        [{'path': '/leases', 'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:lease:get',
        'rule:is_admin or rule:is_lease_owner or rule:is_lease_lessee',
        'Retrieve a single lease',
        [{'path': '/leases/{lease_ident}', 'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:lease:get_all',
        'rule:is_admin or rule:is_owner or rule:is_lessee',
        'Retrieve all leases owned by project_id',
        [{'path': '/leases', 'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:lease:delete',
        'rule:is_admin or rule:is_lease_owner or rule:is_lease_lessee',
        'Delete lease',
        [{'path': '/leases/{lease_ident}', 'method': 'DELETE'}]),
]

offer_policies = [
    policy.DocumentedRuleDefault(
        'esi_leap:offer:offer_admin',
        'rule:is_admin',
        'Complete permissions over offers',
        [{'path': '/offers', 'method': 'POST'},
         {'path': '/offers', 'method': 'GET'},
         {'path': '/offers/{offer_ident}', 'method': 'GET'},
         {'path': '/offers/{offer_ident}', 'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:offer:create',
        'rule:is_admin or rule:is_owner',
        'Create offer',
        [{'path': '/offers', 'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:offer:get',
        'rule:is_admin or rule:is_owner or rule:is_lessee',
        'Retrieve a single offer',
        [{'path': '/offers/{offer_ident}', 'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:offer:get_all',
        'rule:is_admin or rule:is_owner or rule:is_lessee',
        'Retrieve multiple offers',
        [{'path': '/offers', 'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:offer:delete',
        'rule:is_admin or rule:is_offer_owner',
        'Delete offer',
        [{'path': '/offers/{offer_ident}', 'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        'esi_leap:offer:claim',
        'rule:is_admin or rule:is_owner or rule:is_lessee',
        'Claim an offer',
        [{'path': '/offers/{offer_ident}/claim', 'method': 'POST'}]),
]


def list_rules():
    policies = itertools.chain(
        default_policies,
        lease_policies,
        offer_policies,
    )
    return policies


def get_enforcer():
    CONF([], project='esi-leap')
    global _ENFORCER
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(CONF)
        _ENFORCER.register_defaults(list_rules())
    return _ENFORCER


def authorize(rule, target, creds, *args, **kwargs):
    if not CONF.pecan.auth_enable:
        return True

    return get_enforcer().authorize(
        rule, target, creds, do_raise=True, *args, **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from esi_leap.db import api as dbapi


class ResourceObjectInterface(object, metaclass=abc.ABCMeta):
    dbapi = dbapi.get_instance()

    resource_type = 'base'

    @abc.abstractmethod
    def get_uuid(self):
        """Return resource's uuid"""

    @abc.abstractmethod
    def get_name(self, resource_list):
        """Return resource's name, if any"""

    @abc.abstractmethod
    def get_resource_class(self, resource_list):
        """Return resource's associated class, if any"""

    @abc.abstractmethod
    def get_config(self):
        """Return resource's associated config, if any"""

    @abc.abstractmethod
    def get_owner_project_id(self):
        """Return the project id of the resource's owner"""

    @abc.abstractmethod
    def get_lease_uuid(self):
        """Return the uuid of the associated lease, if any"""

    @abc.abstractmethod
    def get_lessee_project_id(self):
        """Return the project id of the associated lessee, if any"""

    @abc.abstractmethod
    def get_node_provision_state(self, resource_list):
        """Return resource's provision state, if any"""

    @abc.abstractmethod
    def get_node_power_state(self, resource_list):
        """Return resource's power state, if any"""

    @abc.abstractmethod
    def set_lease(self, lease):
        """Associates a lease with the resource"""

    @abc.abstractmethod
    def remove_lease(self, lease):
        """Disassociates a lease from the resource"""

    def verify_availability(self, start_time, end_time):
        self.dbapi.resource_verify_availability(
            self.resource_type,
            self.get_uuid(),
            start_time,
            end_time,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.resource_objects import base


class TestNode(base.ResourceObjectInterface):

    resource_type = 'test_node'

    def __init__(self, uuid, project_id='12345'):
        self._uuid = uuid
        self._project_id = project_id

    def get_uuid(self):
        return self._uuid

    def get_name(self, resource_list=None):
        return 'test-node-%s' % self._uuid

    def get_resource_class(self, resource_list=None):
        return 'fake'

    def get_config(self):
        return {}

    def get_owner_project_id(self):
        return self._project_id

    def get_lease_uuid(self):
        return '12345'

    def get_lessee_project_id(self):
        return self._project_id

    def get_node_power_state(self):
        return 'Off'

    def get_node_provision_state(self):
        return 'available'

    def set_lease(self, lease):
        return

    def remove_lease(self, lease):
        return



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from esi_leap.common.exception import ResourceTypeUnknown
from esi_leap.resource_objects import base
# types derived from base won't show as subclasses unless imported somewhere
from esi_leap.resource_objects import dummy_node  # noqa: F401
from esi_leap.resource_objects import ironic_node  # noqa: F401
from esi_leap.resource_objects import test_node  # noqa: F401


_RESOURCE_TYPE_MAP = {
    typ.resource_type: typ for typ in
    base.ResourceObjectInterface.__subclasses__()}
RESOURCE_TYPES = tuple(_RESOURCE_TYPE_MAP.keys())


def get_type(resource_type):
    if resource_type in RESOURCE_TYPES:
        return _RESOURCE_TYPE_MAP[resource_type]
    else:
        raise ResourceTypeUnknown(resource_type=resource_type)


def get_resource_object(resource_type, resource_ident):
    return get_type(resource_type)(resource_ident)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client

from esi_leap.common.i18n import _


class ESILeapException(Exception):
    msg_fmt = _('An unknown exception occurred.')
    code = http_client.INTERNAL_SERVER_ERROR
    safe = False

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                message = self.msg_fmt

        self.message = message
        super(ESILeapException, self).__init__(message)


class LeaseExceedMaxTimeRange(ESILeapException):
    code = http_client.FORBIDDEN
    msg_fmt = _('Attempted to create lease with an invalid '
                'time range. The max time range is %(max_time)s days.')


class HTTPForbidden(ESILeapException):
    code = http_client.FORBIDDEN
    msg_fmt = _('Access was denied to %(rule)s.')


class HTTPResourceForbidden(ESILeapException):
    code = http_client.FORBIDDEN
    msg_fmt = _('Access was denied to %(resource_type)s %(resource)s.')


class LeaseNoPermission(ESILeapException):
    msg_fmt = _('You do not have permissions on '
                'lease %(lease_uuid)s.')


class LeaseDuplicateName(ESILeapException):
    msg_fmt = _('Duplicate leases with name %(name)s.')


class LeaseNotActive(ESILeapException):
    msg_fmt = _('Lease with name or uuid %(lease_id)s not active.')


class LeaseNotFound(ESILeapException):
    code = http_client.NOT_FOUND
    msg_fmt = _('Lease with name or uuid %(lease_id)s not found.')


class LeaseNoOfferUUID(ESILeapException):
    msg_fmt = _('Cannot create lease without parameter offer_uuid.')


class LeaseNoTimeAvailabilities(ESILeapException):
    msg_fmt = _('Lease %(lease_uuid)s has no availabilities at given '
                'time range %(start_time)s, %(end_time)s.')


class OfferNoPermission(ESILeapException):
    msg_fmt = _('You do not have permissions on '
                'offer %(offer_uuid)s.')


class OfferDuplicateName(ESILeapException):
    msg_fmt = _('Duplicate offers with name %(name)s.')


class OfferNotFound(ESILeapException):
    code = http_client.NOT_FOUND
    msg_fmt = _('Offer with name or uuid %(offer_uuid)s not found.')


class OfferNoTimeAvailabilities(ESILeapException):
    msg_fmt = _('Offer %(offer_uuid)s has no availabilities at given '
                'time range %(start_time)s, %(end_time)s.')


class OfferNotAvailable(ESILeapException):
    msg_fmt = _('Offer %(offer_uuid)s does not have status '
                '"available". Got offer status "%(status)s".')


class ProjectNoPermission(ESILeapException):
    msg_fmt = _('You do not have permissions on project %(project_id)s.')


class ProjectNoSuchName(ESILeapException):
    msg_fmt = _('No project named %(name)s.')


class ResourceTimeConflict(ESILeapException):
    msg_fmt = ('Time conflict for %(resource_type)s %(resource_uuid)s.')


class ResourceNoPermission(ESILeapException):
    msg_fmt = _('You do not have permissions on '
                '%(resource_type)s %(resource_uuid)s.')


class ResourceNoPermissionTime(ESILeapException):
    msg_fmt = _('You do not have permissions on '
                '%(resource_type)s %(resource_uuid)s for the time range '
                '%(start_time)s - %(end_time)s.')


class ResourceTypeUnknown(ESILeapException):
    msg_fmt = _('%(resource_type)s resource type unknown.')


class InvalidTimeAPICommand(ESILeapException):
    msg_fmt = _('Attempted to get %(resource)s resource without providing '
                'both a valid Start Time and End Time. '
                'Start Time must be strictly less than End Time. '
                'Got %(start_time)s, %(end_time)s.')


class InvalidAvailabilityAPICommand(ESILeapException):
    msg_fmt = _('Attempted to get an offer resource without providing '
                'both a valid Availability Start Time and Availability '
                'End Time. Availability Start Time must be strictly '
                'less than Availability End Time. '
                'Got %(a_start)s, %(a_end)s.')


class InvalidTimeRange(ESILeapException):
    msg_fmt = _('Attempted to create %(resource)s resource with an invalid '
                'Start Time and End Time. Start Time must be strictly less '
                'than End Time. Got %(start_time)s, %(end_time)s.')


class NodeNotFound(ESILeapException):
    code = http_client.NOT_FOUND
    msg_fmt = _('Encountered an error fetching info for node %(uuid)s '
                '(%(resource_type)s): %(err)s')


class NotificationPayloadError(ESILeapException):
    _msg_fmt = _("Payload not populated when trying to send notification "
                 "\"%(class_name)s\"")


class NotificationSchemaObjectError(ESILeapException):
    _msg_fmt = _("Expected object %(obj)s when populating notification payload"
                 " but got object %(source)s")


class NotificationSchemaKeyError(ESILeapException):
    _msg_fmt = _("Object %(obj)s doesn't have the field \"%(field)s\" "
                 "required for populating notification schema key "
                 "\"%(key)s\"")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.db import api as dbapi
from esi_leap.objects import base
from esi_leap.objects import fields

from oslo_config import cfg
from oslo_log import log as logging
from oslo_versionedobjects import base as versioned_objects_base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@versioned_objects_base.VersionedObjectRegistry.register
class Event(base.ESILEAPObject):
    dbapi = dbapi.get_instance()

    fields = {
        'id': fields.IntegerField(),
        'event_type': fields.StringField(),
        'event_time': fields.DateTimeField(),
        'object_type': fields.StringField(nullable=True),
        'object_uuid': fields.StringField(nullable=True),
        'resource_type': fields.StringField(nullable=True),
        'resource_uuid': fields.StringField(nullable=True),
        'lessee_id': fields.StringField(nullable=True),
        'owner_id': fields.StringField(nullable=True),
    }

    @classmethod
    def get_all(cls, filters, context=None):
        db_events = cls.dbapi.event_get_all(filters)
        return cls._from_db_object_list(context, db_events)

    def create(self, context=None):
        updates = self.obj_get_changes()

        LOG.info('Creating event')
        db_event = self.dbapi.event_create(updates)
        self._from_db_object(context, self, db_event)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json

import mock

from esi_leap.common import statuses
from esi_leap.resource_objects import dummy_node
from esi_leap.tests import base


start = datetime.datetime(2016, 7, 16, 19, 20, 30)


class FakeLease(object):
    def __init__(self):
        self.uuid = '001'
        self.project_id = '654321'
        self.start_time = start + datetime.timedelta(days=5)
        self.end_time = start + datetime.timedelta(days=10)
        self.status = statuses.CREATED
        self.offer_uuid = '534653c9-880d-4c2d-6d6d-f4f2a09e384'


class TestDummyNode(base.TestCase):

    test_node_1 = {
        'project_owner_id': '123456',
        'project_id': '654321',
        'lease_uuid': '001',
        'resource_class': 'fake',
        'power_state': 'off',
        'provision_state': 'enroll',
        'server_config': {
            'new attribute XYZ': 'new attribute XYZ',
            'cpu_type': 'Intel Xeon',
            'cores': 16,
            'ram_gb': 512,
            'storage_type': 'samsung SSD',
            'storage_size_gb': 204
        }
    }

    test_node_2 = {
        'project_owner_id': '123456',
        'resource_class': 'fake',
        'server_config': {
            'new attribute XYZ': 'new attribute XYZ',
            'cpu_type': 'Intel Xeon',
            'cores': 16,
            'ram_gb': 512,
            'storage_type': 'samsung SSD',
            'storage_size_gb': 204
        }
    }

    def setUp(self):
        super(TestDummyNode, self).setUp()
        self.fake_dummy_node = dummy_node.DummyNode('1111')
        self.fake_read_data_1 = json.dumps(self.test_node_1)
        self.fake_read_data_2 = json.dumps(self.test_node_2)

    def test_resource_type(self):
        self.assertEqual('dummy_node', self.fake_dummy_node.resource_type)

    def test_get_uuid(self):
        self.assertEqual('1111', self.fake_dummy_node.get_uuid())

    def test_get_name(self):
        self.assertEqual('dummy-node-1111',
                         self.fake_dummy_node.get_name())

    def test_get_resource_class(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            resource_class = self.fake_dummy_node.get_resource_class()
            self.assertEqual(resource_class,
                             self.test_node_1['resource_class'])
            mock_file_open.assert_called_once()

    def test_get_config(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            config = self.fake_dummy_node.get_config()
            self.assertEqual(config, self.test_node_1['server_config'])
            mock_file_open.assert_called_once()

    def test_get_owner_project_id(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            self.assertEqual(self.fake_dummy_node.get_owner_project_id(),
                             self.test_node_1['project_owner_id'])
            self.assertEqual(mock_file_open.call_count, 1)

    def test_get_lease_uuid(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            lease_uuid = self.fake_dummy_node.get_lease_uuid()
            self.assertEqual(lease_uuid, '001')
            mock_file_open.assert_called_once()

    def test_get_lessee_project_id(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            project_id = self.fake_dummy_node.get_lessee_project_id()
            self.assertEqual(project_id, '654321')
            mock_file_open.assert_called_once()

    def test_get_node_power_state(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            power_state = self.fake_dummy_node.get_node_power_state()
            self.assertEqual(power_state, 'off')
            mock_file_open.assert_called_once()

    def test_get_node_provision_state(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            provision_state = self.fake_dummy_node.get_node_provision_state()
            self.assertEqual(provision_state, 'enroll')
            mock_file_open.assert_called_once()

    def test_set_lease(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_2)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            fake_lease = FakeLease()
            self.fake_dummy_node.set_lease(fake_lease)
            self.assertEqual(mock_file_open.call_count, 2)

    def test_remove_lease(self):
        mock_open = mock.mock_open(read_data=self.fake_read_data_1)
        with mock.patch('builtins.open', mock_open) as mock_file_open:
            fake_lease = FakeLease()
            self.fake_dummy_node.remove_lease(fake_lease)
            self.assertEqual(mock_file_open.call_count, 2)

    @mock.patch('builtins.open')
    def test_get_deleted_node_info(self, mock_open):
        mock_open.side_effect = FileNotFoundError
        self.assertEqual(self.fake_dummy_node.get_resource_class(),
                         'unknown-class')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
import pecan
from pecan import rest

from esi_leap.api.controllers.v1 import root


API_VERSION = 'v1'


class RootController(rest.RestController):

    _versions = [API_VERSION]
    _default_version = API_VERSION

    v1 = root.Controller()

    @pecan.expose(content_type='application/json')
    def index(self):
        pecan.response.status_code = 300
        pecan.response.content_type = 'application/json'
        versions = {
            'versions': [
                {
                    'id': 'v1.0',
                    'status': 'CURRENT',
                    'links': [{
                        'href': '{0}/v1'.format(pecan.request.host_url),
                        'rel': 'self'
                    }]
                }
            ]
        }
        return jsonutils.dump_as_bytes(versions)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os.path

from oslo_log import log as logging

from esi_leap.common import exception
import esi_leap.conf
from esi_leap.resource_objects import base
from esi_leap.resource_objects import error


CONF = esi_leap.conf.CONF
DUMMY_NODE_DIR = CONF.dummy_node.dummy_node_dir

LOG = logging.getLogger(__name__)


class DummyNode(base.ResourceObjectInterface):

    resource_type = 'dummy_node'

    def __init__(self, uuid):
        self._uuid = uuid
        self._path = os.path.join(DUMMY_NODE_DIR, uuid)

    def get_uuid(self):
        return self._uuid

    def get_name(self, resource_list=None):
        return 'dummy-node-%s' % self._uuid

    def get_resource_class(self, resource_list=None):
        return self._get_node_attr('resource_class', '',
                                   resource_list=resource_list,
                                   err_msg='Error getting resource class',
                                   err_val=error.UNKNOWN['resource_class'])

    def get_config(self):
        return self._get_node_attr('server_config', {},
                                   err_msg='Error getting resource config',
                                   err_val=error.UNKNOWN['config'])

    def get_owner_project_id(self):
        return self._get_node_attr('project_owner_id', None,
                                   err_msg='Error getting owner project id',
                                   err_val=error.UNKNOWN['owner_project_id'])

    def get_lease_uuid(self):
        return self._get_node_attr('lease_uuid', '',
                                   err_msg='Error getting lease UUID',
                                   err_val=error.UNKNOWN['lease_uuid'])

    def get_lessee_project_id(self):
        return self._get_node_attr('project_id', '',
                                   err_msg='Error getting lessee project id',
                                   err_val=error.UNKNOWN['lessee_project_id'])

    def get_node_power_state(self):
        return self._get_node_attr('power_state', '',
                                   err_msg='Error getting node power state',
                                   err_val=error.UNKNOWN['power_state'])

    def get_node_provision_state(self):
        return self._get_node_attr('provision_state', '',
                                   err_msg='Error getting'
                                   'node provision state',
                                   err_val=error.UNKNOWN['provision_state'])

    def set_lease(self, lease):
        node_dict = self._get_node()
        node_dict['lease_uuid'] = lease.uuid
        node_dict['project_id'] = lease.project_id
        with open(self._path, 'w') as node_file:
            json.dump(node_dict, node_file)

    def remove_lease(self, lease):
        node_dict = self._get_node()
        node_dict.pop('lease_uuid', None)
        node_dict.pop('project_id', None)
        with open(self._path, 'w') as node_file:
            json.dump(node_dict, node_file)

    def _get_node(self):
        try:
            with open(self._path) as node_file:
                return json.load(node_file)
        except FileNotFoundError as e:
            raise exception.NodeNotFound(uuid=self._uuid,
                                         resource_type=self.resource_type,
                                         err=str(e))

    def _get_node_attr(self, attr, default=None, resource_list=None,
                       err_val=None, err_msg=None):
        try:
            return self._get_node().get(attr, default)
        except exception.NodeNotFound:
            LOG.exception(err_msg)
            return err_val if err_val is not None else default



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_db import api as db_api
from oslo_log import log as logging


_BACKEND_MAPPING = {
    'sqlalchemy': 'esi_leap.db.sqlalchemy.api',
}

IMPL = db_api.DBAPI.from_config(cfg.CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)
LOG = logging.getLogger(__name__)


def get_instance():
    """Return a DB API instance."""
    return IMPL


# Helpers for building constraints / equality checks


def constraint(**conditions):
    """Return a constraint object suitable for use with some updates."""
    return IMPL.constraint(**conditions)


def equal_any(*values):
    """Return an equality condition object suitable for use in a constraint.

    Equal_any conditions require that a model object's attribute equal any
    one of the given values.
    """
    return IMPL.equal_any(*values)


def not_equal(*values):
    """Return an inequality condition object suitable for use in a constraint.

    Not_equal conditions require that a model object's attribute differs from
    all of the given values.
    """
    return IMPL.not_equal(*values)


def to_dict(func):
    def decorator(*args, **kwargs):
        res = func(*args, **kwargs)

        try:
            return [item.to_dict() for item in res]
        except (AttributeError, TypeError):
            if res:
                return res.to_dict()
            else:
                return None

    return decorator


# Offer
@to_dict
def offer_get_by_uuid(offer_uuid):
    return IMPL.offer_get_by_uuid(offer_uuid)


@to_dict
def offer_get_by_name(offer_name):
    return IMPL.offer_get(offer_name)


@to_dict
def offer_get_all():
    return IMPL.offer_get_all()


@to_dict
def offer_get_conflict_times(offer_ref):
    return IMPL.offer_get_conflict_times(offer_ref)


def offer_get_next_lease_start_time(offer_uuid, start):
    return IMPL.offer_get_next_lease_start_time(
        offer_uuid, start)


def offer_verify_availability(offer_ref, start, end):
    return IMPL.offer_verify_availability(
        offer_ref, start, end)


def offer_create(values):
    return IMPL.offer_create(values)


def offer_update(context, offer_uuid, values):
    return IMPL.offer_update(context, offer_uuid, values)


def offer_destroy(offer_uuid):
    return IMPL.offer_destroy(offer_uuid)


# Lease
@to_dict
def lease_get_by_uuid(lease_uuid):
    return IMPL.lease_get(lease_uuid)


@to_dict
def lease_get_by_name(lease_name):
    return IMPL.lease_get(lease_name)


@to_dict
def lease_get_all():
    return IMPL.lease_get_all()


def lease_create(values):
    return IMPL.lease_create(values)


def lease_update(lease_uuid, values):
    return IMPL.lease_update(lease_uuid, values)


def lease_destroy(lease_uuid):
    return IMPL.lease_destroy(lease_uuid)


def lease_verify_child_availability(lease_ref, start, end):
    return IMPL.lease_verify_child_availability(
        lease_ref, start, end)


# Resource object
def resource_verify_availability(r_type, r_uuid, start, end):
    return IMPL.resource_verify_availability(
        r_type, r_uuid, start, end)


def resource_check_admin(resource_type, resource_uuid,
                         start_time, end_time,
                         default_admin_project_id, project_id):
    return IMPL.resource_check_admin(
        resource_type, resource_uuid, start_time, end_time,
        default_admin_project_id, project_id)


# Event
@to_dict
def event_get_all():
    return IMPL.event_get_all()


def event_create(values):
    return IMPL.event_create(values)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.fic language governing permissions and limitations
#    under the License.

from esi_leap.common import exception
from esi_leap.conf import CONF
from oslo_context import context as ctx
import oslo_messaging as messaging
from osprofiler import profiler

NOTIFICATION_TRANSPORT = None
VERSIONED_NOTIFIER = None

ALLOWED_EXMODS = [
    exception.__name__,
]


def init(conf):
    global NOTIFICATION_TRANSPORT
    global VERSIONED_NOTIFIER
    NOTIFICATION_TRANSPORT = messaging.get_notification_transport(
        conf,
        allowed_remote_exmods=ALLOWED_EXMODS)

    serializer = RequestContextSerializer(messaging.JsonPayloadSerializer())

    if conf.notification.notification_level is None:

        VERSIONED_NOTIFIER = messaging.Notifier(NOTIFICATION_TRANSPORT,
                                                serializer=serializer,
                                                driver='noop')
    else:
        VERSIONED_NOTIFIER = messaging.Notifier(
            NOTIFICATION_TRANSPORT,
            serializer=serializer,
            topics=CONF.notification.versioned_notifications_topics)


def cleanup():
    global NOTIFICATION_TRANSPORT
    global VERSIONED_NOTIFIER
    assert NOTIFICATION_TRANSPORT is not None
    assert VERSIONED_NOTIFIER is not None
    NOTIFICATION_TRANSPORT.cleanup()
    NOTIFICATION_TRANSPORT = None
    VERSIONED_NOTIFIER = None


# RequestContextSerializer borrowed from Ironic
class RequestContextSerializer(messaging.Serializer):

    def __init__(self, base):
        self._base = base

    def serialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.serialize_entity(context, entity)

    def deserialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.deserialize_entity(context, entity)

    def serialize_context(self, context):
        _context = context.to_dict()
        prof = profiler.get()
        if prof:
            trace_info = {
                "hmac_key": prof.hmac_key,
                "base_id": prof.get_base_id(),
                "parent_id": prof.get_id()
            }
            _context.update({"trace_info": trace_info})
        return _context

    def deserialize_context(self, context):
        trace_info = context.pop("trace_info", None)
        if trace_info:
            profiler.init(**trace_info)
        return ctx.RequestContext.from_dict(context)


def get_versioned_notifier(publisher_id=None):
    assert VERSIONED_NOTIFIER is not None
    assert publisher_id is not None
    return VERSIONED_NOTIFIER.prepare(publisher_id=publisher_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import six

from oslo_versionedobjects import fields as object_fields


class DateTimeField(object_fields.DateTimeField):
    def __init__(self, **kwargs):
        super(DateTimeField, self).__init__(False, **kwargs)


# FlexibleDict borrowed from Ironic
class FlexibleDict(object_fields.FieldType):
    @staticmethod
    def coerce(obj, attr, value):
        if isinstance(value, six.string_types):
            value = ast.literal_eval(value)
        return dict(value)


class FlexibleDictField(object_fields.AutoTypedField):
    AUTO_TYPE = FlexibleDict()

    def _null(self, obj, attr):
        if self.nullable:
            return {}
        super(FlexibleDictField, self)._null(obj, attr)


class IntegerField(object_fields.IntegerField):
    pass


class ListOfObjectsField(object_fields.ListOfObjectsField):
    pass


class ObjectField(object_fields.ObjectField):
    pass


class StringField(object_fields.StringField):
    pass


class UUIDField(object_fields.UUIDField):
    pass


class NotificationLevel(object_fields.Enum):
    DEBUG = 'debug'
    INFO = 'info'
    WARNING = 'warning'
    ERROR = 'error'
    CRITICAL = 'critical'

    ALL = (DEBUG, INFO, WARNING, ERROR, CRITICAL)

    def __init__(self):
        super(NotificationLevel, self).__init__(
            valid_values=NotificationLevel.ALL)


class NotificationLevelField(object_fields.BaseEnumField):
    AUTO_TYPE = NotificationLevel()


class NotificationStatus(object_fields.Enum):
    START = 'start'
    END = 'end'
    ERROR = 'error'
    SUCCESS = 'success'

    ALL = (START, END, ERROR, SUCCESS)

    def __init__(self):
        super(NotificationStatus, self).__init__(
            valid_values=NotificationStatus.ALL)


class NotificationStatusField(object_fields.BaseEnumField):
    AUTO_TYPE = NotificationStatus()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock
from oslo_utils import uuidutils

from esi_leap.common import exception as e
from esi_leap.common import statuses
from esi_leap.db.sqlalchemy import api
import esi_leap.tests.base as base

now = datetime.datetime(2016, 7, 16, 19, 20, 30)

test_offer_1 = dict(
    uuid='11111',
    project_id='0wn3r',
    name='o1',
    resource_uuid='1111',
    resource_type='dummy_node',
    start_time=now,
    end_time=now + datetime.timedelta(days=100),
    properties={'foo': 'bar'},
    status=statuses.AVAILABLE,
)

test_offer_2 = dict(
    uuid='22222',
    project_id='0wn3r',
    lessee_id='12345',
    name='o1',
    resource_uuid='1111',
    resource_type='dummy_node',
    start_time=now + datetime.timedelta(days=25),
    end_time=now + datetime.timedelta(days=100),
    properties={'foo': 'bar'},
    status=statuses.AVAILABLE,
)

test_offer_3 = dict(
    uuid='33333',
    project_id='0wn3r_2',
    lessee_id='67890',
    name='o1',
    resource_uuid='1111',
    resource_type='dummy_node',
    start_time=now + datetime.timedelta(days=50),
    end_time=now + datetime.timedelta(days=100),
    properties={'foo': 'bar'},
    status=statuses.AVAILABLE,
)

test_offer_4 = dict(
    uuid='44444',
    project_id='0wn3r_2',
    lessee_id='ABCDE',
    name='o2',
    resource_uuid='1111',
    resource_type='dummy_node',
    start_time=now + datetime.timedelta(days=75),
    end_time=now + datetime.timedelta(days=100),
    properties={'foo': 'bar'},
    status=statuses.AVAILABLE,
)

test_offer_5 = dict(
    uuid='55555',
    project_id='12345',
    lessee_id='ABCDE',
    name='o2',
    resource_uuid='1111',
    resource_type='dummy_node',
    start_time=now + datetime.timedelta(days=105),
    end_time=now + datetime.timedelta(days=150),
    properties={'foo': 'bar'},
    status=statuses.AVAILABLE,
)

test_lease_1 = dict(
    uuid='11111',
    project_id='1e5533',
    owner_id='0wn3r',
    name='l1',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=10),
    end_time=now + datetime.timedelta(days=20),
    properties={},
    status=statuses.CREATED,
)

test_lease_2 = dict(
    uuid='22222',
    project_id='1e5533',
    owner_id='0wn3r',
    name='l1',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=20),
    end_time=now + datetime.timedelta(days=30),
    properties={},
    status=statuses.CREATED,
)

test_lease_3 = dict(
    uuid='33333',
    project_id='1e5533_2',
    owner_id='0wn3r_2',
    name='l1',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=50),
    end_time=now + datetime.timedelta(days=60),
    properties={},
    status=statuses.ACTIVE,
)

test_lease_4 = dict(
    uuid='44444',
    project_id='1e5533_2',
    owner_id='0wn3r_2',
    name='l4',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=85),
    end_time=now + datetime.timedelta(days=90),
    properties={},
    status=statuses.DELETED,
)

test_lease_5 = dict(
    project_id='0wn3r',
    owner_id='0wn3r_2',
    name='l5',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=90),
    end_time=now + datetime.timedelta(days=100),
    uuid='55555',
    properties={},
    status=statuses.EXPIRED,
)

test_lease_6 = dict(
    project_id='0wn3r',
    owner_id='0wn3r_2',
    name='l6',
    resource_uuid='2222',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now + datetime.timedelta(days=5),
    end_time=now + datetime.timedelta(days=30),
    uuid='6666',
    properties={},
    status=statuses.EXPIRED,
)

test_lease_7 = dict(
    uuid='77777',
    project_id='1e5533_2',
    owner_id='0wn3r_2',
    name='l1',
    resource_uuid='1111',
    resource_type='dummy_node',
    purpose='test_purpose',
    start_time=now - datetime.timedelta(days=10),
    end_time=now + datetime.timedelta(days=10),
    properties={},
    status=statuses.ACTIVE,
)

test_event_1 = dict(
    id=1,
    event_type='fake:event:start',
    event_time=now - datetime.timedelta(days=20),
    object_type='lease',
    object_uuid='11111',
    resource_type='dummy_node',
    resource_uuid='1111',
    lessee_id='1e5533',
    owner_id='0wn3r',
)

test_event_2 = dict(
    id=2,
    event_type='fake:event:end',
    event_time=now - datetime.timedelta(days=10),
    object_type='lease',
    object_uuid='11111',
    resource_type='dummy_node',
    resource_uuid='1111',
    lessee_id='1e5533',
    owner_id='0wn3r',
)

test_event_3 = dict(
    id=3,
    event_type='fake:event:start',
    event_time=now - datetime.timedelta(days=20),
    object_type='lease',
    object_uuid='22222',
    resource_type='dummy_node',
    resource_uuid='2222',
    lessee_id='1e5533',
    owner_id='0wn3r2',
)


class TestOfferAPI(base.DBTestCase):

    def test_offer_create(self):
        offer = api.offer_create(test_offer_1)
        o = api.offer_get_all({}).all()
        assert len(o) == 1
        assert o[0].to_dict() == offer.to_dict()

    def test_offer_verify_availability(self):
        offer = api.offer_create(test_offer_1)

        test_lease_1['offer_uuid'] = offer.uuid
        test_lease_2['offer_uuid'] = offer.uuid
        test_lease_3['offer_uuid'] = offer.uuid

        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)
        api.lease_create(test_lease_3)

        start = now + datetime.timedelta(days=35)
        end = now + datetime.timedelta(days=40)
        api.offer_verify_availability(offer, start, end)

        start = now + datetime.timedelta(days=5)
        end = now + datetime.timedelta(days=10)
        api.offer_verify_availability(offer, start, end)

        start = now
        end = now + datetime.timedelta(days=10)
        api.offer_verify_availability(offer, start, end)

        start = now + datetime.timedelta(days=90)
        end = now + datetime.timedelta(days=100)
        api.offer_verify_availability(offer, start, end)

        start = now + datetime.timedelta(days=60)
        end = now + datetime.timedelta(days=100)
        api.offer_verify_availability(offer, start, end)

        start = now + datetime.timedelta(days=30)
        end = now + datetime.timedelta(days=50)
        api.offer_verify_availability(offer, start, end)

        start = now + datetime.timedelta(days=15)
        end = now + datetime.timedelta(days=16)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=45)
        end = now + datetime.timedelta(days=55)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=55)
        end = now + datetime.timedelta(days=65)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=50)
        end = now + datetime.timedelta(days=65)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=45)
        end = now + datetime.timedelta(days=60)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=90)
        end = now + datetime.timedelta(days=105)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=100)
        end = now + datetime.timedelta(days=105)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=105)
        end = now + datetime.timedelta(days=110)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now - datetime.timedelta(days=1)
        end = now + datetime.timedelta(days=5)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now - datetime.timedelta(days=1)
        end = now
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now - datetime.timedelta(days=10)
        end = now - datetime.timedelta(days=5)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        start = now + datetime.timedelta(days=45)
        end = now + datetime.timedelta(days=55)
        self.assertRaises(e.OfferNoTimeAvailabilities,
                          api.offer_verify_availability,
                          offer, start, end)

        test_lease_4['offer_uuid'] = offer.uuid
        api.lease_create(test_lease_4)
        start = now + datetime.timedelta(days=86)
        end = now + datetime.timedelta(days=87)
        api.offer_verify_availability(offer, start, end)

    def test_offer_get_conflict_times(self):
        o1 = api.offer_create(test_offer_1)
        self.assertEqual(api.offer_get_conflict_times(o1), [])
        test_lease_3['offer_uuid'] = o1.uuid
        test_lease_4['offer_uuid'] = o1.uuid
        test_lease_5['offer_uuid'] = o1.uuid
        api.lease_create(test_lease_3)
        api.lease_create(test_lease_4)
        api.lease_create(test_lease_5)
        self.assertEqual(api.offer_get_conflict_times(o1),
                         [(now + datetime.timedelta(days=50),
                          now + datetime.timedelta(days=60))])

    def test_offer_get_next_lease_start_time(self):
        o1 = api.offer_create(test_offer_1)
        self.assertEqual(api.offer_get_next_lease_start_time
                         (o1.uuid, o1.start_time,), None)
        test_lease_3['offer_uuid'] = o1.uuid
        api.lease_create(test_lease_3)
        test_lease_7['offer_uuid'] = o1.uuid
        api.lease_create(test_lease_7)
        self.assertEqual(
            api.offer_get_next_lease_start_time(
                o1.uuid, o1.start_time),
            (now + datetime.timedelta(days=50),))

    def test_offer_get_by_uuid(self):
        o1 = api.offer_create(test_offer_1)
        res = api.offer_get_by_uuid(o1.uuid)
        self.assertEqual(o1.uuid, res.uuid)
        self.assertEqual(o1.project_id, res.project_id)
        self.assertEqual(o1.properties, res.properties)

    def test_offer_get_by_uuid_not_found(self):
        assert api.offer_get_by_uuid('some_uuid') is None

    def test_offer_get_by_name(self):
        o1 = api.offer_create(test_offer_1)
        o2 = api.offer_create(test_offer_2)
        o3 = api.offer_create(test_offer_3)
        api.offer_create(test_offer_4)

        res = api.offer_get_by_name('o1')
        assert len(res) == 3
        self.assertEqual(o1.uuid, res[0].uuid)
        self.assertEqual(o1.project_id, res[0].project_id)

        self.assertEqual(o2.uuid, res[1].uuid)
        self.assertEqual(o2.project_id, res[1].project_id)

        self.assertEqual(o3.uuid, res[2].uuid)
        self.assertEqual(o3.project_id, res[2].project_id)

    def test_offer_get_by_name_not_found(self):
        self.assertEqual(api.offer_get_by_name('some_name'), [])

    def test_offer_destroy(self):
        o1 = api.offer_create(test_offer_2)
        api.offer_destroy(o1.uuid)
        self.assertEqual(api.offer_get_by_uuid('offer_2'), None)

    def test_offer_destroy_not_found(self):
        self.assertEqual(api.offer_get_by_uuid('offer_4'), None)

    def test_offer_update(self):
        o1 = api.offer_create(test_offer_3)
        values = {'start_time': test_offer_2['start_time'],
                  'end_time': test_offer_2['end_time']}
        api.offer_update(o1.uuid, values)
        o1 = api.offer_get_by_uuid(o1.uuid)

        self.assertEqual(test_offer_2['start_time'], o1.start_time)
        self.assertEqual(test_offer_2['end_time'], o1.end_time)

    def test_offer_update_invalid_time(self):
        o1 = api.offer_create(test_offer_3)
        values = {'start_time': now + datetime.timedelta(days=101),
                  'end_time': now}
        self.assertRaises(e.InvalidTimeRange, api.offer_update,
                          o1.uuid, values)

    def test_offer_get_all(self):
        o1 = api.offer_create(test_offer_2)
        o2 = api.offer_create(test_offer_3)
        res = api.offer_get_all({})

        self.assertEqual(2, res.count())
        self.assertEqual((o1.to_dict(), o2.to_dict()),
                         (res[0].to_dict(), res[1].to_dict()))

    @mock.patch('esi_leap.common.keystone.get_parent_project_id_tree')
    def test_offer_get_all_lessee_filter(self, mock_gppit):
        mock_gppit.return_value = ['12345', '67890']

        o1 = api.offer_create(test_offer_1)
        o2 = api.offer_create(test_offer_2)
        o3 = api.offer_create(test_offer_3)
        api.offer_create(test_offer_4)
        o5 = api.offer_create(test_offer_5)
        res = api.offer_get_all({'lessee_id': '12345'})

        mock_gppit.assert_called_once_with('12345')
        self.assertEqual(4, res.count())
        self.assertEqual((o1.to_dict(), o2.to_dict(), o3.to_dict(),
                          o5.to_dict()),
                         (res[0].to_dict(), res[1].to_dict(),
                          res[2].to_dict(), res[3].to_dict()))

    def test_offer_get_all_time_filter(self):
        o1 = api.offer_create(test_offer_1)
        o2 = api.offer_create(test_offer_2)
        api.offer_create(test_offer_3)
        api.offer_create(test_offer_4)
        api.offer_create(test_offer_5)
        res = api.offer_get_all({
            'start_time': o1.start_time + datetime.timedelta(days=26),
            'end_time': o1.end_time + datetime.timedelta(days=-1),
        })

        self.assertEqual(2, res.count())
        self.assertEqual((o1.to_dict(), o2.to_dict()),
                         (res[0].to_dict(), res[1].to_dict()))

    def test_offer_get_all_time_filter_within(self):
        o1 = api.offer_create(test_offer_1)
        o2 = api.offer_create(test_offer_2)
        o3 = api.offer_create(test_offer_3)
        o4 = api.offer_create(test_offer_4)
        api.offer_create(test_offer_5)
        res = api.offer_get_all({
            'start_time': o1.end_time + datetime.timedelta(days=-2),
            'end_time': o2.end_time + datetime.timedelta(days=1),
            'time_filter_type': 'within'
        })

        self.assertEqual(4, res.count())
        self.assertEqual((o1.to_dict(), o2.to_dict(), o3.to_dict(),
                          o4.to_dict()),
                         (res[0].to_dict(), res[1].to_dict(),
                          res[2].to_dict(), res[3].to_dict()))


class TestLeaseAPI(base.DBTestCase):

    def test_lease_get_by_uuid(self):
        o1 = api.offer_create(test_offer_2)
        test_lease_4['offer_uuid'] = o1.uuid
        l1 = api.lease_create(test_lease_4)
        res = api.lease_get_by_uuid(l1.uuid)
        self.assertEqual(l1.uuid, res.uuid)

    def test_lease_get_by_uuid_not_found(self):
        assert api.lease_get_by_uuid('some_uuid') is None

    def test_lease_get_by_name(self):
        o1 = api.offer_create(test_offer_1)
        o2 = api.offer_create(test_offer_2)
        o3 = api.offer_create(test_offer_3)
        o4 = api.offer_create(test_offer_4)
        test_lease_1['offer_uuid'] = o1.uuid
        test_lease_2['offer_uuid'] = o2.uuid
        test_lease_3['offer_uuid'] = o3.uuid
        test_lease_4['offer_uuid'] = o4.uuid
        l1 = api.lease_create(test_lease_1)
        l2 = api.lease_create(test_lease_2)
        l3 = api.lease_create(test_lease_3)
        api.lease_create(test_lease_4)
        res = api.lease_get_by_name('l1')
        assert len(res) == 3
        self.assertEqual(l1.uuid, res[0].uuid)
        self.assertEqual(l1.project_id, res[0].project_id)

        self.assertEqual(l2.uuid, res[1].uuid)
        self.assertEqual(l2.project_id, res[1].project_id)

        self.assertEqual(l3.uuid, res[2].uuid)
        self.assertEqual(l3.project_id, res[2].project_id)

    def test_lease_get_by_name_not_found(self):
        self.assertEqual(api.lease_get_by_name('some_name'), [])

    def test_lease_get_all(self):
        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)

        res = api.lease_get_all({})
        res_uuids = [res_lease.to_dict()['uuid'] for res_lease in res]

        self.assertEqual(2, res.count())
        self.assertIn(test_lease_1['uuid'], res_uuids)
        self.assertIn(test_lease_2['uuid'], res_uuids)

    def test_lease_get_all_filter_by_status(self):
        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)
        api.lease_create(test_lease_3)
        api.lease_create(test_lease_4)
        api.lease_create(test_lease_5)

        res = api.lease_get_all({'status': [statuses.CREATED,
                                            statuses.ACTIVE]})
        res_uuids = [res_lease.to_dict()['uuid'] for res_lease in res]

        self.assertEqual(3, res.count())
        self.assertIn(test_lease_1['uuid'], res_uuids)
        self.assertIn(test_lease_2['uuid'], res_uuids)
        self.assertIn(test_lease_3['uuid'], res_uuids)

    def test_lease_get_all_filter_by_time(self):
        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)
        api.lease_create(test_lease_3)

        start_time = test_lease_2['start_time'] + datetime.timedelta(days=1)
        end_time = test_lease_2['end_time'] + datetime.timedelta(days=-1)

        res = api.lease_get_all({'start_time': start_time,
                                 'end_time': end_time})
        res_uuids = [res_lease.to_dict()['uuid'] for res_lease in res]

        self.assertEqual(1, res.count())
        self.assertIn(test_lease_2['uuid'], res_uuids)

    def test_lease_get_all_filter_by_time_within(self):
        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)
        api.lease_create(test_lease_3)
        api.lease_create(test_lease_6)

        start_time = test_lease_1['end_time'] + datetime.timedelta(days=-1)
        end_time = test_lease_2['start_time'] + datetime.timedelta(days=1)

        res = api.lease_get_all({'start_time': start_time,
                                 'end_time': end_time,
                                 'time_filter_type': 'within'})
        res_uuids = [res_lease.to_dict()['uuid'] for res_lease in res]

        self.assertEqual(3, res.count())
        self.assertIn(test_lease_1['uuid'], res_uuids)
        self.assertIn(test_lease_2['uuid'], res_uuids)
        self.assertIn(test_lease_6['uuid'], res_uuids)

    def test_lease_get_all_filter_by_project_or_owner_id(self):
        api.lease_create(test_lease_1)
        api.lease_create(test_lease_2)
        api.lease_create(test_lease_3)
        api.lease_create(test_lease_4)
        api.lease_create(test_lease_5)

        res = api.lease_get_all({'project_or_owner_id': '0wn3r'})
        res_uuids = [res_lease.to_dict()['uuid'] for res_lease in res]

        self.assertEqual(3, res.count())
        self.assertIn(test_lease_1['uuid'], res_uuids)
        self.assertIn(test_lease_2['uuid'], res_uuids)
        self.assertIn(test_lease_5['uuid'], res_uuids)

    def test_lease_create(self):
        o1 = api.offer_create(test_offer_2)
        test_lease_4['offer_uuid'] = o1.uuid
        l1 = api.lease_create(test_lease_4)
        l2 = api.lease_get_all({}).all()
        assert len(l2) == 1
        assert l2[0].to_dict() == l1.to_dict()

    def test_lease_update(self):
        o1 = api.offer_create(test_offer_2)
        test_lease_4['offer_uuid'] = o1.uuid
        l1 = api.lease_create(test_lease_4)
        values = {'start_time': test_lease_5['start_time'],
                  'end_time': test_lease_5['end_time']}
        api.lease_update(l1.uuid, values)
        l1 = api.lease_get_by_uuid(l1.uuid)
        self.assertEqual(test_lease_5['start_time'], l1.start_time)
        self.assertEqual(test_lease_5['end_time'], l1.end_time)

    def test_lease_update_invalid_time(self):
        o1 = api.offer_create(test_offer_3)
        test_lease_4['offer_uuid'] = o1.uuid
        l1 = api.lease_create(test_lease_4)
        values = {'start_time': now + datetime.timedelta(days=101),
                  'end_time': now}
        self.assertRaises(e.InvalidTimeRange, api.lease_update,
                          l1.uuid, values)

    def test_lease_destroy(self):
        o1 = api.offer_create(test_offer_2)
        test_lease_4['offer_uuid'] = o1.uuid
        l1 = api.lease_create(test_lease_4)
        api.lease_destroy(l1.uuid)
        self.assertEqual(api.lease_get_by_uuid('lease_4'), None)

    def test_lease_destroy_not_found(self):
        self.assertEqual(api.lease_get_by_uuid('lease_4'), None)


class TestLeaseVerifyChildAvailability(base.DBTestCase):

    def setUp(self):
        super(TestLeaseVerifyChildAvailability, self).setUp()

        self.parent_lease_data = dict(
            uuid=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(),
            owner_id=uuidutils.generate_uuid(),
            resource_uuid='1111',
            resource_type='dummy_node',
            purpose='test_purpose',
            start_time=now + datetime.timedelta(days=50),
            end_time=now + datetime.timedelta(days=100),
            status=statuses.ACTIVE,
        )
        self.child_lease_data = dict(
            uuid=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(),
            owner_id=uuidutils.generate_uuid(),
            parent_lease_uuid=self.parent_lease_data['uuid'],
            resource_uuid='1111',
            resource_type='dummy_node',
            purpose='test_purpose',
            start_time=now + datetime.timedelta(days=60),
            end_time=now + datetime.timedelta(days=70),
            status=statuses.ACTIVE,
        )
        self.child_offer_data = dict(
            uuid=uuidutils.generate_uuid(),
            project_id=uuidutils.generate_uuid(),
            parent_lease_uuid=self.parent_lease_data['uuid'],
            resource_uuid='1111',
            resource_type='dummy_node',
            start_time=now + datetime.timedelta(days=70),
            end_time=now + datetime.timedelta(days=80),
            status=statuses.AVAILABLE,
        )

    def test_lease_verify_child_availability(self):
        parent_lease = api.lease_create(self.parent_lease_data)

        start = (self.parent_lease_data['start_time'] +
                 datetime.timedelta(days=1))
        end = self.parent_lease_data['end_time'] + datetime.timedelta(days=-1)
        api.lease_verify_child_availability(parent_lease, start, end)

        start = (self.parent_lease_data['start_time'] +
                 datetime.timedelta(days=-1))
        end = self.parent_lease_data['end_time'] + datetime.timedelta(days=-1)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

        start = (self.parent_lease_data['start_time'] +
                 datetime.timedelta(days=1))
        end = self.parent_lease_data['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

        start = (self.parent_lease_data['start_time'] +
                 datetime.timedelta(days=-1))
        end = self.parent_lease_data['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

    def test_lease_verify_child_availability_lease_conflict(self):
        parent_lease = api.lease_create(self.parent_lease_data)
        api.lease_create(self.child_lease_data)

        start = (self.child_lease_data['start_time'] +
                 datetime.timedelta(days=-5))
        end = self.child_lease_data['start_time'] + datetime.timedelta(days=-1)
        api.lease_verify_child_availability(parent_lease, start, end)

        start = self.child_lease_data['end_time'] + datetime.timedelta(days=1)
        end = self.child_lease_data['end_time'] + datetime.timedelta(days=5)
        api.lease_verify_child_availability(parent_lease, start, end)

        start = (self.child_lease_data['start_time'] +
                 datetime.timedelta(days=1))
        end = self.child_lease_data['end_time'] + datetime.timedelta(days=5)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

        start = (self.child_lease_data['start_time'] +
                 datetime.timedelta(days=-1))
        end = self.child_lease_data['end_time'] + datetime.timedelta(days=-1)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

    def test_lease_verify_child_availability_offer_conflict(self):
        parent_lease = api.lease_create(self.parent_lease_data)
        api.offer_create(self.child_offer_data)

        start = (self.child_offer_data['start_time'] +
                 datetime.timedelta(days=-5))
        end = self.child_offer_data['start_time'] + datetime.timedelta(days=-1)
        api.lease_verify_child_availability(parent_lease, start, end)

        start = self.child_offer_data['end_time'] + datetime.timedelta(days=1)
        end = self.child_offer_data['end_time'] + datetime.timedelta(days=5)
        api.lease_verify_child_availability(parent_lease, start, end)

        start = (self.child_offer_data['start_time'] +
                 datetime.timedelta(days=1))
        end = self.child_offer_data['end_time'] + datetime.timedelta(days=5)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)

        start = (self.child_offer_data['start_time'] +
                 datetime.timedelta(days=-1))
        end = self.child_offer_data['end_time'] + datetime.timedelta(days=-1)
        self.assertRaises(e.LeaseNoTimeAvailabilities,
                          api.lease_verify_child_availability,
                          parent_lease, start, end,)


class TestResourceVerifyAvailabilityAPI(base.DBTestCase):

    def test_resource_verify_availability_offer_conflict(self):
        o1 = api.offer_create(test_offer_4)
        r_type = o1.resource_type
        r_uuid = o1.resource_uuid

        start = test_offer_4['end_time'] + datetime.timedelta(days=1)
        end = test_offer_4['end_time'] + datetime.timedelta(days=5)
        api.resource_verify_availability(r_type, r_uuid, start, end)

        start = test_offer_4['start_time'] + datetime.timedelta(days=1)
        end = test_offer_4['end_time'] + datetime.timedelta(days=-1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_offer_4['start_time'] + datetime.timedelta(days=-1)
        end = test_offer_4['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_offer_4['start_time'] + datetime.timedelta(days=-1)
        end = test_offer_4['start_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_offer_4['end_time'] + datetime.timedelta(days=-1)
        end = test_offer_4['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

    def test_resource_verify_availability_lease_conflict(self):
        test_lease = api.lease_create(test_lease_1)
        r_type = test_lease.resource_type
        r_uuid = test_lease.resource_uuid

        start = test_lease_1['end_time'] + datetime.timedelta(days=1)
        end = test_lease_1['end_time'] + datetime.timedelta(days=5)
        api.resource_verify_availability(r_type, r_uuid, start, end)

        start = test_lease_1['start_time'] + datetime.timedelta(days=1)
        end = test_lease_1['end_time'] + datetime.timedelta(days=-1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_lease_1['start_time'] + datetime.timedelta(days=-1)
        end = test_lease_1['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_lease_1['start_time'] + datetime.timedelta(days=-1)
        end = test_lease_1['start_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)

        start = test_lease_1['end_time'] + datetime.timedelta(days=-1)
        end = test_lease_1['end_time'] + datetime.timedelta(days=1)
        self.assertRaises(e.ResourceTimeConflict,
                          api.resource_verify_availability,
                          r_type, r_uuid, start, end)


class TestEventAPI(base.DBTestCase):

    def test_event_get_all(self):
        api.event_create(test_event_1)
        api.event_create(test_event_2)

        events = api.event_get_all({})
        event_ids = [event.to_dict()['id'] for event in events]

        self.assertEqual(2, events.count())
        self.assertIn(test_event_1['id'], event_ids)
        self.assertIn(test_event_2['id'], event_ids)

    def test_event_get_all_filter_by_last_event_time(self):
        api.event_create(test_event_1)
        api.event_create(test_event_2)
        api.event_create(test_event_3)

        events = api.event_get_all({'last_event_time':
                                    test_event_1['event_time']})
        event_ids = [event.to_dict()['id'] for event in events]

        self.assertEqual(1, events.count())
        self.assertIn(test_event_2['id'], event_ids)

    def test_event_get_all_filter_by_last_event_id(self):
        api.event_create(test_event_1)
        api.event_create(test_event_2)
        api.event_create(test_event_3)

        events = api.event_get_all({'last_event_id': 1})
        event_ids = [event.to_dict()['id'] for event in events]

        self.assertEqual(2, events.count())
        self.assertIn(test_event_2['id'], event_ids)
        self.assertIn(test_event_3['id'], event_ids)

    def test_event_get_all_filter_by_lessee_or_owner_id(self):
        api.event_create(test_event_1)
        api.event_create(test_event_2)
        api.event_create(test_event_3)

        events = api.event_get_all({'lessee_or_owner_id': '0wn3r'})
        event_ids = [event.to_dict()['id'] for event in events]

        self.assertEqual(2, events.count())
        self.assertIn(test_event_1['id'], event_ids)
        self.assertIn(test_event_2['id'], event_ids)

    def test_lease_create(self):
        event = api.event_create(test_event_1)
        events = api.event_get_all({}).all()
        assert len(events) == 1
        assert events[0].to_dict() == event.to_dict()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from esi_leap.api.controllers import types as types
import mock as mock
import unittest as unittest
from wsme import types as wtypes


class TestCollection(unittest.TestCase):

    def setUp(self):
        self.test_collection = types.Collection()
        self.test_collection._type = 'stuff'
        self.obj1 = mock.Mock(uuid='aaaaa')
        self.obj2 = mock.Mock(uuid='bbbbb')
        self.obj3 = mock.Mock(uuid='ccccc')
        self.test_collection.stuff = [self.obj1, self.obj2, self.obj3]
        self._type = 'url'

    def test_has_next(self):
        self.assertEqual(self.test_collection.has_next(3), True)
        self.assertEqual(self.test_collection.has_next(0), False)

    def test_get_next(self):
        kwargs = {'key1': 'Things', 'key2': 'Stuff'}
        link = ("{{'key1': 'Things', 'key2': 'Stuff'}}"
                "/v1/stuff?limit=3&marker={0}".format(self.obj3.uuid))
        self.assertEqual(self.test_collection.get_next(2, kwargs),
                         wtypes.Unset)
        self.assertEqual(self.test_collection.get_next(3, kwargs), link)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from ironicclient.common.apiclient import exceptions as ir_exception
import mock

from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.resource_objects import ironic_node
from esi_leap.tests import base

start = datetime.datetime(2016, 7, 16, 19, 20, 30)
fake_uuid = '13921c8d-ce11-4b6d-99ed-10e19d184e5f'


class FakeIronicNode(object):
    def __init__(self):
        self.created_at = start
        self.lessee = 'abcdef'
        self.owner = '123456'
        self.name = 'fake-node'
        self.properties = {'lease_uuid': '001'}
        self.provision_state = 'available'
        self.uuid = fake_uuid
        self.resource_class = 'baremetal'
        self.power_state = 'off'


class FakeLease(object):
    def __init__(self):
        self.uuid = '001'
        self.project_id = '654321'
        self.start_time = start + datetime.timedelta(days=5)
        self.end_time = start + datetime.timedelta(days=10)
        self.status = statuses.ACTIVE
        self.offer_uuid = '534653c9-880d-4c2d-6d6d-f4f2a09e384'


class TestIronicNode(base.TestCase):

    def test_resource_type(self):
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        self.assertEqual('ironic_node', test_ironic_node.resource_type)

    def test_get_uuid(self):
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        self.assertEqual(fake_uuid, test_ironic_node.get_uuid())

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_name(self, mock_gn):
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node

        self.assertEqual(fake_get_node.name, test_ironic_node.get_name())
        mock_gn.assert_called_once()

    @mock.patch.object(ironic_node, 'get_ironic_client', autospec=True)
    def test_init_with_name(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value.node.get.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode('node-name')

        self.assertEqual(fake_uuid, test_ironic_node.get_uuid())
        mock_gn.assert_called_once()
        mock_gn.return_value.node.get.assert_called_once_with('node-name')

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_resource_class(self, mock_gn):
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node

        resource_class = test_ironic_node.get_resource_class()
        self.assertEqual('baremetal', resource_class)
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_config(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        config = test_ironic_node.get_config()
        expected_config = fake_get_node.properties
        self.assertEqual(config, expected_config)
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_owner_project_id(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node.get_owner_project_id(),
                         fake_get_node.owner)
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_lease_uuid(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node.get_lease_uuid(),
                         fake_get_node.properties.get('lease_uuid'))
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_lessee_project_id(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node.get_lessee_project_id(),
                         fake_get_node.lessee)
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_node_power_state(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node.get_node_power_state(),
                         fake_get_node.power_state)
        mock_gn.assert_called_once()

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    def test_get_node_provision_state(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node.get_node_provision_state(),
                         fake_get_node.provision_state)
        mock_gn.assert_called_once()

    @mock.patch.object(ironic_node, 'get_ironic_client', autospec=True)
    def test_set_lease(self, client_mock):
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        fake_lease = FakeLease()

        test_ironic_node.set_lease(fake_lease)
        client_mock.assert_called_once()
        client_mock.return_value.node.update.assert_called_once_with(
            fake_uuid, [{'op': 'add',
                         'path': '/properties/lease_uuid',
                         'value': fake_lease.uuid},
                        {'op': 'add',
                         'path': '/lessee',
                         'value': fake_lease.project_id}])

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode.'
                'get_lessee_project_id')
    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode.'
                'get_lease_uuid')
    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode._get_node')
    @mock.patch.object(ironic_node, 'get_ironic_client', autospec=True)
    def test_remove_lease(self, mock_client, mock_gn, mock_glu, mock_glpi):
        fake_get_node = FakeIronicNode()
        fake_get_node.provision_state = 'active'
        fake_lease = FakeLease()

        mock_gn.return_value = fake_get_node
        mock_glu.return_value = fake_lease.uuid
        mock_glpi.return_value = fake_lease.project_id

        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        test_ironic_node.remove_lease(fake_lease)

        mock_glpi.assert_called_once()
        mock_glu.assert_called_once()
        self.assertEqual(mock_gn.call_count, 1)
        self.assertEqual(mock_client.call_count, 2)
        mock_client.return_value.node.update.assert_called_once_with(
            fake_uuid, [{'op': 'remove', 'path': '/properties/lease_uuid'},
                        {'op': 'remove', 'path': '/lessee'}])
        mock_client.return_value.node.set_provision_state. \
            assert_called_once_with(fake_uuid, 'deleted')

    @mock.patch('esi_leap.resource_objects.ironic_node.IronicNode.'
                'get_lease_uuid')
    def test_expire_lease_no_match(self, mock_glu):
        mock_glu.return_value = 'none'
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        fake_lease = FakeLease()

        test_ironic_node.remove_lease(fake_lease)
        mock_glu.assert_called_once()

    @mock.patch('esi_leap.common.ironic.get_node')
    def test_get_node(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)

        self.assertEqual(test_ironic_node._get_node(), fake_get_node)
        mock_gn.assert_called_once_with(fake_uuid, None)

    @mock.patch('esi_leap.common.ironic.get_node')
    def test_get_node_cache(self, mock_gn):
        fake_get_node = FakeIronicNode()
        mock_gn.return_value = fake_get_node
        test_ironic_node = ironic_node.IronicNode(fake_uuid)
        test_ironic_node._node = fake_get_node

        self.assertEqual(fake_get_node,
                         test_ironic_node._get_node())
        mock_gn.assert_not_called()

    @mock.patch('esi_leap.common.ironic.get_node')
    def test_get_unknown_node(self, mock_gn):
        mock_gn.side_effect = ir_exception.NotFound
        test_unknown_node = ironic_node.IronicNode(fake_uuid)

        self.assertRaises(exception.NodeNotFound,
                          test_unknown_node._get_node)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystonemiddleware import auth_token
from oslo_context import context
import pecan
from pecan import hooks

import esi_leap.conf


CONF = esi_leap.conf.CONF


class ContextHook(hooks.PecanHook):
    def before(self, state):
        ctx = context.RequestContext.from_environ(state.request.environ)
        state.request.context = ctx

    def after(self, state):
        state.request.context = None


def get_pecan_config():
    cfg_dict = {
        'app': {
            'root': CONF.pecan.root,
            'modules': CONF.pecan.modules,
            'debug': CONF.pecan.debug,
            'auth_enable': CONF.pecan.auth_enable
        }
    }

    return pecan.configuration.conf_from_dict(cfg_dict)


def setup_app(config=None):
    if not config:
        config = get_pecan_config()

    pecan.configuration.set_config(dict(config), overwrite=True)

    app = pecan.make_app(
        config.app.root,
        hooks=lambda: [ContextHook()],
        debug=CONF.pecan.debug,
        static_root=config.app.static_root if CONF.pecan.debug else None,
        force_canonical=getattr(config.app, 'force_canonical', True),
    )

    if CONF.pecan.auth_enable:
        app = auth_token.AuthProtocol(app, dict(CONF.keystone_authtoken))

    return app


class WSGIApplication(object):
    def __init__(self):
        pecan_cfg = get_pecan_config()
        self.app = setup_app(pecan_cfg)

    def __call__(self, environ, start_response):
        return self.app(environ, start_response)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from esi_leap.common import statuses
from esi_leap.resource_objects import test_node
from esi_leap.tests import base

start = datetime.datetime(2016, 7, 16, 19, 20, 30)


def get_test_lease():
    return {
        'id': 12345,
        'name': 'c',
        'uuid': '534653c9-880d-4c2d-6d6d-11111111111',
        'project_id': 'le55ee',
        'start_time': start + datetime.timedelta(days=5),
        'end_time': start + datetime.timedelta(days=10),
        'fulfill_time': start + datetime.timedelta(days=5),
        'expire_time': start + datetime.timedelta(days=10),
        'status': statuses.CREATED,
        'properties': {},
        'offer_uuid': '534653c9-880d-4c2d-6d6d-f4f2a09e384',
        'created_at': None,
        'updated_at': None
    }


class TestTestNode(base.TestCase):

    def setUp(self):
        super(TestTestNode, self).setUp()
        self.fake_test_node = test_node.TestNode('1111', '123456')

    def test_resource_type(self):
        resource_type = self.fake_test_node.resource_type
        self.assertEqual(resource_type, 'test_node')

    def test_get_uuid(self):
        self.assertEqual(self.fake_test_node.get_uuid(), '1111')

    def test_get_name(self):
        self.assertEqual(self.fake_test_node.get_name(), 'test-node-1111')

    def test_get_resource_class(self):
        self.assertEqual(self.fake_test_node.get_resource_class(), 'fake')

    def test_get_config(self):
        self.assertEqual(self.fake_test_node.get_config(), {})

    def test_get_owner_project_id(self):
        self.assertEqual(self.fake_test_node.get_owner_project_id(), '123456')

    def test_get_lease_uuid(self):
        self.assertEqual(self.fake_test_node.get_lease_uuid(), '12345')

    def test_get_lessee_project_id(self):
        self.assertEqual(self.fake_test_node.get_lessee_project_id(), '123456')

    def test_set_lease(self):
        fake_lease = get_test_lease()
        self.assertEqual(self.fake_test_node.set_lease(fake_lease), None)

    def test_remove_lease(self):
        fake_lease = get_test_lease()
        self.assertEqual(self.fake_test_node.remove_lease(fake_lease), None)

    def test_get_node_power_state(self):
        self.assertEqual(self.fake_test_node.get_node_power_state(), 'Off')

    def test_get_node_provision_state(self):
        self.assertEqual(self.fake_test_node.get_node_provision_state(),
                         'available')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy as oslo_policy

from esi_leap.common import policy
import esi_leap.conf
from esi_leap.tests import base


CONF = esi_leap.conf.CONF


class PolicyTestCase(base.TestCase):

    def setUp(self):
        super(PolicyTestCase, self).setUp()

        CONF.set_override('auth_enable', True,
                          group='pecan')

    def test_authorized(self):
        creds = {'roles': ['esi_leap_owner']}
        self.assertTrue(policy.authorize('esi_leap:offer:get',
                                         creds, creds))

    def test_unauthorized(self):
        creds = {'roles': ['generic_user']}
        self.assertRaises(
            oslo_policy.PolicyNotAuthorized,
            policy.authorize, 'esi_leap:offer:get', creds, creds)

    def test_authorize_policy_not_registered(self):
        creds = {'roles': ['generic_user']}
        self.assertRaises(
            oslo_policy.PolicyNotRegistered,
            policy.authorize, 'esi_leap:foo:bar', creds, creds)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log
from oslo_versionedobjects import base as object_base

from esi_leap.objects import fields as object_fields


LOG = log.getLogger(__name__)


class ESILEAPObject(object_base.VersionedObject):
    OBJ_SERIAL_NAMESPACE = 'esi_leap_object'
    OBJ_PROJECT_NAMESPACE = 'esi_leap'

    fields = {
        'created_at': object_fields.DateTimeField(nullable=True),
        'updated_at': object_fields.DateTimeField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, obj, db_obj):
        for key in obj.fields:
            setattr(obj, key, db_obj[key])
            obj.obj_reset_changes()
        obj._context = context
        return obj

    @classmethod
    def _from_db_object_list(cls, context, db_objs):
        return [cls._from_db_object(context, cls(), db_obj)
                for db_obj in db_objs]

    def to_dict(self):
        return dict((k, getattr(self, k))
                    for k in self.fields
                    if self.obj_attr_is_set(k))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import http.client as http_client
import mock
from oslo_context import context as ctx
from oslo_utils import uuidutils
import testtools

from esi_leap.api.controllers.v1.lease import LeasesController
from esi_leap.common import constants
from esi_leap.common import exception
from esi_leap.common import statuses
from esi_leap.objects import lease as lease_obj
from esi_leap.resource_objects.ironic_node import IronicNode
from esi_leap.resource_objects.test_node import TestNode
from esi_leap.tests.api import base as test_api_base


class TestLeasesController(test_api_base.APITestCase):

    def setUp(self):
        super(TestLeasesController, self).setUp()

        self.test_lease = lease_obj.Lease(
            start_time=datetime.datetime(2016, 7, 16, 19, 20, 30),
            end_time=datetime.datetime(2016, 8, 16, 19, 20, 30),
            uuid=uuidutils.generate_uuid(),
            resource_type='test_node',
            resource_uuid='111',
            project_id='lesseeid',
            owner_id='ownerid',
            parent_lease_uuid=None
        )
        self.test_lease_1 = lease_obj.Lease(
            start_time=datetime.datetime(2016, 7, 16, 19, 20, 30),
            end_time=datetime.datetime(2016, 8, 16, 19, 20, 30),
            uuid=uuidutils.generate_uuid(),
            resource_type='ironic_node',
            resource_uuid='222',
            project_id='lesseeid',
            owner_id='ownerid',
            parent_lease_uuid=None
        )
        self.test_lease_with_parent = lease_obj.Lease(
            start_time=datetime.datetime(2016, 7, 16, 19, 20, 30),
            end_time=datetime.datetime(2016, 8, 16, 19, 20, 30),
            uuid=uuidutils.generate_uuid(),
            resource_type='test_node',
            resource_uuid='111',
            project_id='lesseeid',
            owner_id='ownerid',
            parent_lease_uuid='parent-lease-uuid'
        )

    def test_empty(self):
        data = self.get_json('/leases')
        self.assertEqual([], data['leases'])

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_one(self, mock_ga, mock_lgdwai, mock_gpl, mock_gnl):
        mock_ga.return_value = [self.test_lease]
        mock_lgdwai.return_value = self.test_lease.to_dict()
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        data = self.get_json('/leases')

        self.assertEqual(self.test_lease.uuid,
                         data['leases'][0]['uuid'])
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        mock_lgdwai.assert_called_once()

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    def test_post(self, mock_create, mock_cra, mock_generate_uuid,
                  mock_gpufi, mock_gro, mock_lgdwai):
        resource = TestNode('1234567890')
        data = {
            'project_id': 'lesseeid',
            'resource_type': 'test_node',
            'resource_uuid': '1234567890',
            'start_time': '2016-07-16T19:20:30',
            'end_time': '2016-08-16T19:20:30',
            'purpose': 'test_purpose'
        }
        return_data = data.copy()
        return_data['owner_id'] = self.context.project_id
        return_data['uuid'] = self.test_lease.uuid
        lgdwai_return_data = return_data.copy()
        lgdwai_return_data['start_time'] = datetime.datetime(
            2016, 7, 16, 19, 20, 30)
        lgdwai_return_data['end_time'] = datetime.datetime(
            2016, 8, 16, 19, 20, 30)

        mock_gro.return_value = resource
        mock_gpufi.return_value = 'lesseeid'
        mock_generate_uuid.return_value = self.test_lease.uuid
        mock_lgdwai.return_value = lgdwai_return_data

        request = self.post_json('/leases', data)

        mock_gro.assert_called_once_with('test_node', '1234567890')
        mock_generate_uuid.assert_called_once()
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_create.assert_called_once()
        mock_lgdwai.assert_called_once()
        self.assertEqual(return_data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    def test_post_default_resource_type(self, mock_create, mock_cra,
                                        mock_generate_uuid, mock_gpufi,
                                        mock_gro, mock_lgdwai):
        resource = IronicNode('13921c8d-ce11-4b6d-99ed-10e19d184e5f')
        data = {
            'project_id': 'lesseeid',
            'resource_uuid': '1234567890',
            'start_time': '2016-07-16T19:20:30',
            'end_time': '2016-08-16T19:20:30'
        }
        return_data = data.copy()
        return_data['resource_type'] = 'ironic_node'
        return_data['owner_id'] = self.context.project_id
        return_data['uuid'] = self.test_lease.uuid
        lgdwai_return_data = return_data.copy()
        lgdwai_return_data['start_time'] = datetime.datetime(
            2016, 7, 16, 19, 20, 30)
        lgdwai_return_data['end_time'] = datetime.datetime(
            2016, 8, 16, 19, 20, 30)

        mock_gro.return_value = resource
        mock_gpufi.return_value = 'lesseeid'
        mock_generate_uuid.return_value = self.test_lease.uuid
        mock_lgdwai.return_value = lgdwai_return_data

        request = self.post_json('/leases', data)

        mock_gro.assert_called_once_with('ironic_node', '1234567890')
        mock_generate_uuid.assert_called_once()
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_create.assert_called_once()
        mock_lgdwai.assert_called_once()
        self.assertEqual(return_data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_resource_lease_admin')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    def test_post_non_admin_parent_lease(self, mock_create, mock_cra,
                                         mock_generate_uuid, mock_gpufi,
                                         mock_gro, mock_crla, mock_lgdwai):
        resource = IronicNode('13921c8d-ce11-4b6d-99ed-10e19d184e5f')
        data = {
            'project_id': 'lesseeid',
            'resource_uuid': '1234567890',
            'start_time': '2016-07-17T19:20:30',
            'end_time': '2016-08-14T19:20:30'
        }
        return_data = data.copy()
        return_data['owner_id'] = self.context.project_id
        return_data['uuid'] = self.test_lease_with_parent.uuid
        return_data['resource_type'] = 'ironic_node'
        return_data['parent_lease_uuid'] = (
            self.test_lease_with_parent.parent_lease_uuid)
        lgdwai_return_data = return_data.copy()
        lgdwai_return_data['start_time'] = datetime.datetime(
            2016, 7, 17, 19, 20, 30)
        lgdwai_return_data['end_time'] = datetime.datetime(
            2016, 8, 14, 19, 20, 30)

        mock_gro.return_value = resource
        mock_gpufi.return_value = 'lesseeid'
        mock_generate_uuid.return_value = self.test_lease_with_parent.uuid
        mock_cra.side_effect = exception.HTTPResourceForbidden(
            resource_type='ironic_node', resource='1234567890')
        mock_crla.return_value = self.test_lease_with_parent.parent_lease_uuid
        mock_lgdwai.return_value = lgdwai_return_data

        request = self.post_json('/leases', data)

        mock_gro.assert_called_once_with('ironic_node', '1234567890')
        mock_generate_uuid.assert_called_once()
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_crla.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id,
            datetime.datetime(2016, 7, 17, 19, 20, 30),
            datetime.datetime(2016, 8, 14, 19, 20, 30))
        mock_create.assert_called_once()
        mock_lgdwai.assert_called_once()
        self.assertEqual(return_data, request.json)
        self.assertEqual(http_client.CREATED, request.status_int)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_resource_lease_admin')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch('esi_leap.api.controllers.v1.utils.check_resource_admin')
    @mock.patch('esi_leap.objects.lease.Lease.create')
    def test_post_non_admin_no_parent_lease(self, mock_create, mock_cra,
                                            mock_generate_uuid, mock_gpufi,
                                            mock_gro, mock_crla):
        fake_uuid = '13921c8d-ce11-4b6d-99ed-10e19d184e5f'
        resource = IronicNode(fake_uuid)
        mock_gro.return_value = resource
        mock_gpufi.return_value = 'lesseeid'
        mock_generate_uuid.return_value = self.test_lease.uuid
        mock_cra.side_effect = exception.HTTPResourceForbidden(
            resource_type='ironic_node', resource=fake_uuid)
        mock_crla.return_value = None

        data = {
            'project_id': 'lesseeid',
            'resource_uuid': fake_uuid,
            'start_time': '2016-07-17T19:20:30',
            'end_time': '2016-08-14T19:20:30'
        }
        request = self.post_json('/leases', data, expect_errors=True)

        mock_gro.assert_called_once_with('ironic_node', fake_uuid)
        mock_generate_uuid.assert_called_once()
        mock_cra.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id)
        mock_crla.assert_called_once_with(
            self.context.to_policy_values(),
            resource,
            self.context.project_id,
            datetime.datetime(2016, 7, 17, 19, 20, 30),
            datetime.datetime(2016, 8, 14, 19, 20, 30))
        mock_create.assert_not_called()
        self.assertEqual(http_client.FORBIDDEN, request.status_int)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_nofilters(self, mock_get_all, mock_lgaaf, mock_lgdwai,
                           mock_gpl, mock_gnl):
        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        self.get_json('/leases')

        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id=None,
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id=None,
                                           resource_type=None,
                                           resource_uuid=None)
        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_project_filter(self, mock_get_all, mock_lgaaf,
                                mock_gpufi, mock_lgdwai, mock_gpl,
                                mock_gnl):
        mock_gpufi.return_value = '12345'
        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []
        self.get_json('/leases?project_id=12345')

        mock_gpufi.assert_called_once_with('12345')
        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id='12345',
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id=None,
                                           resource_type=None,
                                           resource_uuid=None)
        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.common.keystone.get_project_uuid_from_ident')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_owner_filter(self, mock_get_all, mock_lgaaf,
                              mock_gpufi, mock_lgdwai, mock_gpl,
                              mock_gnl):
        mock_gpufi.return_value = '54321'
        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        self.get_json('/leases?owner_id=54321')

        mock_gpufi.assert_called_once_with('54321')
        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id=None,
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id='54321',
                                           resource_type=None,
                                           resource_uuid=None)

        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_resource_filter(self, mock_get_all, mock_lgaaf,
                                 mock_gro, mock_lgdwai, mock_gpl,
                                 mock_gnl):
        mock_gro.return_value = TestNode('54321')
        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        self.get_json('/leases?resource_uuid=54321&resource_type=test_node')

        mock_gro.assert_called_once_with('test_node', '54321')
        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id=None,
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id=None,
                                           resource_type='test_node',
                                           resource_uuid='54321')

        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_resource_class_filter(self, mock_get_all, mock_lgaaf,
                                       mock_lgdwai, mock_gpl, mock_gnl):
        def _get_lease_response(l, use_datetime=False):
            if use_datetime:
                start = datetime.datetime(2016, 7, 16, 19, 20, 30)
                end = datetime.datetime(2016, 8, 16, 19, 20, 30)
            else:
                start = '2016-07-16T19:20:30'
                end = '2016-08-16T19:20:30'

            if l.resource_type in ['test_node', 'dummy_node']:
                resource_class = 'fake'
            elif l.resource_type == 'ironic_node':
                resource_class = 'baremetal'

            return {
                'resource_type': l.resource_type,
                'resource_uuid': l.resource_uuid,
                'resource_class': resource_class,
                'project_id': l.project_id,
                'start_time': start,
                'end_time': end,
                'uuid': l.uuid,
                'owner_id': l.owner_id,
                'parent_lease_uuid': None
            }

        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []
        mock_lgdwai.side_effect = [_get_lease_response(self.test_lease,
                                                       use_datetime=True),
                                   _get_lease_response(self.test_lease_1,
                                                       use_datetime=True)]
        response = self.get_json('/leases?resource_class=fake')

        expected_resp = {'leases': [_get_lease_response(self.test_lease)]}

        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id=None,
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id=None,
                                           resource_type=None,
                                           resource_uuid=None)

        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)
        self.assertEqual(response, expected_resp)

    @mock.patch('esi_leap.common.ironic.get_node_list')
    @mock.patch('esi_leap.common.keystone.get_project_list')
    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'lease_get_dict_with_added_info')
    @mock.patch('esi_leap.api.controllers.v1.lease.get_resource_object')
    @mock.patch('esi_leap.api.controllers.v1.lease.LeasesController.'
                '_lease_get_all_authorize_filters')
    @mock.patch('esi_leap.objects.lease.Lease.get_all')
    def test_get_resource_filter_default_resource_type(self, mock_get_all,
                                                       mock_lgaaf, mock_gro,
                                                       mock_lgdwai, mock_gpl,
                                                       mock_gnl):
        fake_uuid = uuidutils.generate_uuid()
        mock_gro.return_value = IronicNode(fake_uuid)
        mock_get_all.return_value = [self.test_lease, self.test_lease]
        mock_gpl.return_value = []
        mock_gnl.return_value = []

        self.get_json('/leases?resource_uuid=%s' % fake_uuid)

        mock_gro.assert_called_once_with('ironic_node', fake_uuid)
        mock_lgaaf.assert_called_once_with(self.context.to_policy_values(),
                                           project_id=None,
                                           start_time=None,
                                           end_time=None,
                                           status=None,
                                           offer_uuid=None,
                                           view=None,
                                           owner_id=None,
                                           resource_type='ironic_node',
                                           resource_uuid=fake_uuid)

        mock_get_all.assert_called_once()
        mock_gpl.assert_called_once()
        mock_gnl.assert_called_once()
        self.assertEqual(2, mock_lgdwai.call_count)

    @mock.patch('esi_leap.api.controllers.v1.utils.'
                'check_lease_policy_and_retrieve')
    @mock.patch('esi_leap.objects.lease.Lease.cancel')
    def test_lease_delete(self, mock_cancel, mock_clpar):
        mock_clpar.return_value = self.test_lease

        self.delete_json('/leases/' + self.test_lease.uuid)

        mock_clpar.assert_called_once_with(self.context,
                                           'esi_leap:lease:get',
                                           self.test_lease.uuid,
                                           statuses.LEASE_CAN_DELETE)
        mock_cancel.assert_called_once()


class TestLeaseControllersGetAllFilters(testtools.TestCase):

    def setUp(self):
        super(TestLeaseControllersGetAllFilters, self).setUp()

        self.admin_ctx = ctx.RequestContext(project_id='adminid',
                                            roles=['admin'])
        self.owner_ctx = ctx.RequestContext(project_id='ownerid',
                                            roles=['owner'])
        self.lessee_ctx = ctx.RequestContext(project_id='lesseeid',
                                             roles=['lessee'])
        self.random_ctx = ctx.RequestContext(project_id='randomid',
                                             roles=['randomrole'])

    def test_lease_get_all_no_view_no_projectid_no_owner(self):

        expected_filters = {
            'status': ['random'],
            'offer_uuid': 'offeruuid',
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        expected_filters['project_or_owner_id'] = self.admin_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(),
            status='random', offer_uuid='offeruuid')
        self.assertEqual(expected_filters, filters)

        # owner
        expected_filters['project_or_owner_id'] = self.owner_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.owner_ctx.to_policy_values(),
            status='random', offer_uuid='offeruuid')
        self.assertEqual(expected_filters, filters)

        # lessee
        expected_filters['project_or_owner_id'] = self.lessee_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.lessee_ctx.to_policy_values(),
            status='random', offer_uuid='offeruuid')
        self.assertEqual(expected_filters, filters)

        # random
        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.random_ctx.to_policy_values(),
                          status='random', offer_uuid='offeruuid')

    def test_lease_get_all_no_view_project_no_owner(self):

        expected_filters = {
            'status': ['random'],
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        expected_filters['project_id'] = self.admin_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(),
            project_id=self.admin_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        # lessee
        expected_filters['project_id'] = self.lessee_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.lessee_ctx.to_policy_values(),
            project_id=self.lessee_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.random_ctx.to_policy_values(),
                          project_id=self.random_ctx.project_id,
                          status='random')

        # owner
        expected_filters['project_id'] = self.owner_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.owner_ctx.to_policy_values(),
            project_id=self.owner_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        # random
        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.random_ctx.to_policy_values(),
                          project_id=self.random_ctx.project_id,
                          status='random')

    def test_lease_get_all_no_view_any_projectid_owner(self):

        expected_filters = {
            'status': ['random'],
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        expected_filters['owner_id'] = self.admin_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(),
            owner_id=self.admin_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        # lessee
        expected_filters['owner_id'] = self.lessee_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.lessee_ctx.to_policy_values(),
            owner_id=self.lessee_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        expected_filters['owner_id'] = self.lessee_ctx.project_id
        expected_filters['project_id'] = self.random_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.lessee_ctx.to_policy_values(),
            owner_id=self.lessee_ctx.project_id,
            project_id=self.random_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)
        del expected_filters['project_id']

        # owner
        expected_filters['owner_id'] = self.owner_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.owner_ctx.to_policy_values(),
            owner_id=self.owner_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        expected_filters['owner_id'] = self.owner_ctx.project_id
        expected_filters['project_id'] = self.random_ctx.project_id
        filters = LeasesController._lease_get_all_authorize_filters(
            self.owner_ctx.to_policy_values(),
            owner_id=self.owner_ctx.project_id,
            project_id=self.random_ctx.project_id,
            status='random')
        self.assertEqual(expected_filters, filters)

        # random
        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.random_ctx.to_policy_values(),
                          owner_id=self.random_ctx.project_id,
                          status='random')

    def test_lease_get_all_all_view(self):

        expected_filters = {
            'status': ['random'],
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(),
            view='all',
            status='random')
        self.assertEqual(expected_filters, filters)

        # not admin
        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.lessee_ctx.to_policy_values(),
                          view='all',
                          status='random')

        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.owner_ctx.to_policy_values(),
                          view='all',
                          status='random')

        self.assertRaises(exception.HTTPForbidden,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.random_ctx.to_policy_values(),
                          view='all',
                          status='random')

    def test_lease_get_all_all_view_times(self):

        start = datetime.datetime(2016, 7, 16, 19, 20, 30)
        end = datetime.datetime(2020, 7, 16, 19, 20, 30)

        expected_filters = {
            'status': ['random'],
            'start_time': start,
            'end_time': end,
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(),
            view='all', start_time=start, end_time=end, status='random')
        self.assertEqual(expected_filters, filters)

        self.assertRaises(exception.InvalidTimeAPICommand,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.admin_ctx.to_policy_values(),
                          view='all', start_time=start, status='random')

        self.assertRaises(exception.InvalidTimeAPICommand,
                          LeasesController.
                          _lease_get_all_authorize_filters,
                          self.admin_ctx.to_policy_values(),
                          view='all', end_time=end, status='random')

    def test_lease_get_all_status(self):

        expected_filters = {
            'project_or_owner_id': 'adminid',
            'status': [statuses.CREATED, statuses.ACTIVE, statuses.ERROR,
                       statuses.WAIT_CANCEL, statuses.WAIT_EXPIRE,
                       statuses.WAIT_FULFILL],
            'time_filter_type': constants.WITHIN_TIME_FILTER
        }

        # admin
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values())
        self.assertEqual(expected_filters, filters)

        del(expected_filters['status'])
        filters = LeasesController._lease_get_all_authorize_filters(
            self.admin_ctx.to_policy_values(), status='any')
        self.assertEqual(expected_filters, filters)
